from .condition import ConditionModelSerializer, ConditionRepresentationSerializer
from .data import (
    DataModelSerializer,
    DataRepresentationSerializer,
    DataValueModelSerializer,
    DataValueRepresentationSerializer,
)
from .display import DisplayModelSerializer, DisplayRepresentationSerializer
from .process import (
    ProcessModelSerializer,
    ProcessRepresentationSerializer,
    ProcessStepModelSerializer,
    ProcessStepRepresentationSerializer,
    AssignedProcessStepSerializer,
)
from .step import (
    DecisionStepModelSerializer,
    DecisionStepRepresentationSerializer,
    EmailStepModelSerializer,
    EmailStepRepresentationSerializer,
    FinishStepModelSerializer,
    FinishStepRepresentationSerializer,
    JoinStepModelSerializer,
    JoinStepRepresentationSerializer,
    ScriptStepModelSerializer,
    ScriptStepRepresentationSerializer,
    SplitStepModelSerializer,
    SplitStepRepresentationSerializer,
    StepModelSerializer,
    StepRepresentationSerializer,
    UserStepModelSerializer,
    UserStepRepresentationSerializer,
    StartStepModelSerializer,
    StartStepRepresentationSerializer,
)
from .transition import TransitionModelSerializer, TransitionRepresentationSerializer
from .workflow import WorkflowModelSerializer, WorkflowRepresentationSerializer



# Generated by Django 5.0.14 on 2025-05-21 11:05

from django.conf import settings
from django.db import migrations, models
from django.db.models import Count


def handle_duplicates(apps, schema_editor):
    NotificationTypeSetting = apps.get_model("notifications", "NotificationTypeSetting")
    for row in NotificationTypeSetting.objects.values('user', 'notification_type').annotate(c=Count("*")).filter(c__gt=1):
        duplicates = NotificationTypeSetting.objects.filter(user=row['user'], notification_type=row['notification_type'])
        for duplicate in duplicates[1:]:
            duplicate.delete()

class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0008_notificationtype_is_lock'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RunPython(handle_duplicates),
        migrations.AlterModelOptions(
            name='notificationtypesetting',
            options={'verbose_name': 'Notification Type Setting', 'verbose_name_plural': 'Notification Type Settings'},
        ),
        migrations.AddConstraint(
            model_name='notificationtypesetting',
            constraint=models.UniqueConstraint(fields=('user', 'notification_type'), name='unique_notification_setting'),
        ),
    ]



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

BUILDING_MENUITEM = MenuItem(
    label=_("Buildings"),
    endpoint="wbcore:agenda:building-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["agenda.view_building"],
    ),
    add=MenuItem(
        label=_("Create Building"),
        endpoint="wbcore:agenda:building-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["agenda.add_building"],
        ),
    ),
)

CONFERENCE_ROOM_MENUITEM = MenuItem(
    label=_("Conference Rooms"),
    endpoint="wbcore:agenda:conferenceroom-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["agenda.view_conferenceroom"],
    ),
    add=MenuItem(
        label=_("Create Conference Room"),
        endpoint="wbcore:agenda:conferenceroom-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["agenda.add_conferenceroom"],
        ),
    ),
)



from django.contrib import admin
from mptt.admin import MPTTModelAdmin

from .models import Geography


@admin.register(Geography)
class GeographyAdmin(MPTTModelAdmin):
    search_fields = ("name", "code_2", "code_3")
    list_display = ("name", "code_2", "code_3", "parent")
    raw_id_fields = ["parent"]
    readonly_fields = ["level"]
    fieldsets = (
        (
            "Main information",
            {
                "fields": (
                    ("name", "short_name", "alternative_names", "representation"),
                    ("code_2", "code_3", "code"),
                    ("population", "ranking", "time_zone"),
                    (
                        "parent",
                        "level",
                    ),
                    ("search_vector",),
                )
            },
        ),
    )
    list_filter = ("level",)



class Style:
    """A convinience class to have all styles and helper methods contained in one place.

    Static CSS Variables may be used as classmembers, e.g.: `Style.MIN_CONTENT` or `Style.AUTO`
    Dynamic CSS Variables, such as specific sizes, may be used as static methods, e.g.: `Style.fr(5)`
    """

    MIN_CONTENT = "min-content"
    MAX_CONTENT = "max-content"
    AUTO = "auto"

    @staticmethod
    def fr(fractions: int) -> str:
        return Style._stringify(fractions, "fr")

    @staticmethod
    def px(pixels: int) -> str:
        return Style._stringify(pixels, "px")

    @staticmethod
    def pct(percent: float) -> str:
        return Style._stringify(percent, "%")

    @staticmethod
    def rem(rem: float) -> str:
        return Style._stringify(rem, "rem")

    @staticmethod
    def vh(vh: int) -> str:
        return Style._stringify(vh, "vh")

    @staticmethod
    def vw(vw: int) -> str:
        return Style._stringify(vw, "vw")

    @staticmethod
    def _stringify(value: int | float, unit: str) -> str:
        return f"{value}{unit}"

    @staticmethod
    def minmax(min_value: str, max_value: str) -> str:
        return f"minmax({min_value}, {max_value})"



import random

import factory
from django.utils.timezone import make_aware
from faker import Faker
from pandas.tseries.offsets import BDay
from psycopg.types.range import TimestamptzRange

from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.icons.icons import WBIcon

from .conference_room import ConferenceRoomFactory

fake = Faker()


def _get_random_period():
    lower = Faker().future_datetime() + BDay(0)
    upper = lower + BDay(random.randint(1, 7))
    return TimestamptzRange(
        lower=make_aware(lower),
        upper=make_aware(upper),
    )


class CalendarItemFactory(factory.django.DjangoModelFactory):
    color = factory.Faker("color")
    icon = factory.LazyAttribute(lambda x: WBIcon[random.choice(WBIcon.names)].icon)
    visibility = CalendarItem.Visibility.PUBLIC
    period = factory.LazyAttribute(lambda x: _get_random_period())
    title = factory.Faker("sentence")
    conference_room = factory.SubFactory(ConferenceRoomFactory)
    all_day = False
    is_cancelled = False
    is_deletable = True

    @factory.post_generation
    def entities(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for entitie in extracted:
                self.entities.add(entitie)

    class Meta:
        model = CalendarItem
        django_get_or_create = ("title",)



from .condition import CONDITION_MENUITEM
from .data import DATA_MENUITEM
from .process import PROCESS_MENUITEM, PROCESSSTEP_MENUITEM
from .step import (
    DECISIONSTEP_MENUITEM,
    EMAILSTEP_MENUITEM,
    FINISHSTEP_MENUITEM,
    JOINSTEP_MENUITEM,
    SCRIPTSTEP_MENUITEM,
    SPLITSTEP_MENUITEM,
    USERSTEP_MENUITEM,
    STARTSTEP_MENUITEM,
)
from .transition import TRANSITION_MENUITEM
from .workflow import WORKFLOW_MENUITEM



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers
from wbcore import shares
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs.buttons import ActionButton, ButtonConfig
from wbcore.metadata.configs.display.instance_display import (
    create_simple_display,
    create_simple_section,
    repeat_field,
)

from .sites import share_site


@shares.register(
    section=create_simple_section(
        "base_section",
        "Share to someone",
        [[repeat_field(2, "share"), repeat_field(2, "share_recipients")], [repeat_field(4, "share_message")]],
        collapsed=True,
    ),
    weight=-100,
)
class DefaultShareSerializer(wb_serializers.Serializer):
    share = wb_serializers.BooleanField(label=_("Share to someone"), default=False)
    share_message = wb_serializers.TextField(
        label=_("Message"),
        default="Check out this Widget.",
        required=False,
        depends_on=[{"field": "share", "options": {}}],
    )
    share_recipients = wb_serializers.PrimaryKeyRelatedField(
        many=True,
        queryset=get_user_model().objects.all(),
        label=_("Recipient"),
        required=False,
        depends_on=[{"field": "share", "options": {}}],
    )
    _share_recipients = UserRepresentationSerializer(
        many=True, source="share_recipients", depends_on=[{"field": "share", "options": {}}]
    )


def share_action_button(request):
    fields = [[repeat_field(4, "widget_endpoint")]]
    for section in share_site.sorted_sections:
        fields.append([repeat_field(4, section.key)])

    btn = ActionButton(
        icon=WBIcon.SHARE.icon,
        description_fields=_("Are you sure you want to share this widget?"),
        endpoint=reverse("wbcore:share", request=request),
        instance_display=create_simple_display(fields, share_site.sorted_sections),
        confirm_config=ButtonConfig(label=_("Share"), icon=WBIcon.SHARE.icon),
        serializer=share_site.serializer_class,
        action_label=_("Share"),
        title=_("Share"),
        label=_("Share"),
    )
    return btn



from datetime import date, datetime, time

import pytest
import pytz
from django.test import override_settings
from rest_framework.exceptions import ValidationError

from wbcore.serializers import DateField, DateTimeField, TimeField
from wbcore.serializers.fields.types import WBCoreType


class TestDateTimeField:
    def setup_method(self):
        self.field = DateTimeField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize(
        "input, expected",
        [
            ("2019-01-01T10:00", datetime(2019, 1, 1, 11, 0)),
            ("2019-01-01T10:00Z", datetime(2019, 1, 1, 11, 0)),
            ("2019-01-01T10:00+0100", datetime(2019, 1, 1, 10, 0)),
            ("2019-01-01T10:00:00Z", datetime(2019, 1, 1, 11, 0)),
            ("2019-01-01T10:00:00+0000", datetime(2019, 1, 1, 11, 0)),
            ("2019-01-01T10:00:00+0100", datetime(2019, 1, 1, 10, 0)),
            ("2019-01-01T10:00:00.0000Z", datetime(2019, 1, 1, 11, 0)),
            ("2019-01-01T10:00:00.0000+0100", datetime(2019, 1, 1, 10, 0)),
        ],
    )
    @override_settings(TIME_ZONE="UCT", USE_TZ=True)
    def test_to_internal_value(self, input, expected):
        expected = pytz.timezone("Europe/Berlin").localize(expected)
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", ["", "200-00-10", [], {}, None])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    @override_settings(TIME_ZONE="Europe/Berlin")
    def test_to_representation_non_utc(self):
        localized_dt = pytz.timezone("Europe/Berlin").localize(datetime(2019, 1, 1, 10, 0))
        assert self.field.to_representation(localized_dt) == "2019-01-01T10:00:00+0100"

    @override_settings(TIME_ZONE="UCT", USE_TZ=True)
    def test_to_representation_utc(self):
        localized_dt = pytz.timezone("UCT").localize(datetime(2019, 1, 1, 10, 0))
        assert self.field.to_representation(localized_dt) == "2019-01-01T10:00:00+0000"

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.DATETIME.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )


class TestDateField:
    def setup_method(self):
        self.field = DateField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize(
        "input, expected",
        [(date(2019, 1, 1), date(2019, 1, 1)), ("2019-01-01", date(2019, 1, 1))],
    )
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", ["", "200-00-10", [], {}, None])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    def test_to_representation(self):
        assert self.field.to_representation(date(2019, 1, 1)) == "2019-01-01"

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.DATE.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )


class TestTimeField:
    def setup_method(self):
        self.field = TimeField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize(
        "input, expected",
        [
            (time(10, 0), time(10, 0)),
            ("10:00", time(10, 0)),
            ("10:00:00", time(10, 0)),
            ("10:00:00.0000", time(10, 0)),
        ],
    )
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", ["", "111", [], {}, None])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    def test_to_representation(self):
        assert self.field.to_representation(time(10, 0, 0, 0)) == "10:00:00"

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.TIME.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "decorators": [],
                "depends_on": [],
            },
        )



from unittest.mock import MagicMock

import pytest
from django.db import models
from django_fsm import TransitionNotAllowed
from dynamic_preferences.registries import global_preferences_registry
from pytest_mock import MockerFixture

from wbcore.contrib.authentication.factories import InternalUserFactory
from wbcore.contrib.directory.factories import ClientManagerRelationshipFactory as CMRF
from wbcore.contrib.directory.factories import (
    CompanyFactory,
    EmailContactFactory,
    EntryFactory,
    PersonFactory,
    TelephoneContactFactory,
)
from wbcore.contrib.directory.models import ClientManagerRelationship as CMR
from wbcore.contrib.directory.models import (
    Company,
    EmailContact,
    Entry,
    Person,
    RelationshipType,
    TelephoneContact,
)
from wbcore.contrib.directory.models.entries import handle_user_deactivation


@pytest.fixture
def test_person():
    return PersonFactory()


@pytest.mark.directory_model_tests
class TestSpecificModelContacts:
    def refresh_from_db(self, model_entries: list[models.Model]) -> None:
        for entry in model_entries:
            entry.refresh_from_db()

    @pytest.fixture
    def test_entry(self):
        return EntryFactory()

    @pytest.fixture
    def test_company(self):
        return CompanyFactory()

    @pytest.mark.with_db
    @pytest.mark.django_db
    @pytest.mark.parametrize("is_primary, has_duplicates", [(a, b) for a in [True, False] for b in [True, False]])
    def test_set_entry_primary_telephone(self, is_primary, has_duplicates, test_entry):
        # Arrange
        telephone_contact_a = TelephoneContactFactory(entry=test_entry, primary=is_primary)
        telephone_contact_b = TelephoneContactFactory(entry=test_entry if has_duplicates else None)
        # Act
        TelephoneContact.set_entry_primary_telephone(test_entry, telephone_contact_b.number)
        self.refresh_from_db([telephone_contact_a, telephone_contact_b])
        # Assert
        if has_duplicates:
            assert test_entry.telephones.get(primary=True) == telephone_contact_b
            assert not telephone_contact_a.primary
            assert telephone_contact_b.primary
        elif is_primary:
            assert test_entry.telephones.get(primary=True).number == telephone_contact_b.number
        else:
            number_b = telephone_contact_b.number
            assert TelephoneContact.objects.filter(entry=test_entry, number=number_b, primary=True).exists()

    @pytest.mark.with_db
    @pytest.mark.django_db
    @pytest.mark.parametrize("is_primary, has_duplicates", [(a, b) for a in [True, False] for b in [True, False]])
    def test_set_entry_primary_email(self, is_primary, has_duplicates, test_entry):
        # Arrange
        email_contact_a = EmailContactFactory(entry=test_entry, primary=is_primary)
        email_contact_b = EmailContactFactory(entry=test_entry if has_duplicates else None)
        # Act
        EmailContact.set_entry_primary_email(test_entry, email_contact_b.address)
        self.refresh_from_db([email_contact_a, email_contact_b])
        # Assert
        if has_duplicates:
            assert test_entry.emails.get(primary=True) == email_contact_b
            assert not email_contact_a.primary
            assert email_contact_b.primary
        elif is_primary:
            assert test_entry.emails.get(primary=True).address == email_contact_b.address
        else:
            address_b = email_contact_b.address
            assert EmailContact.objects.filter(entry=test_entry, address=address_b, primary=True).exists()

    @pytest.mark.with_db
    @pytest.mark.django_db
    @pytest.mark.parametrize("with_contact", (True, False))
    def test_get_primary_contact_with_db(self, with_contact: bool, test_entry):
        contact = EmailContactFactory(entry=test_entry if with_contact else None, primary=True)
        result = test_entry._primary_contact(test_entry.emails)
        assert result == contact if with_contact else result is None

    @pytest.mark.parametrize(
        "contact_method, contact_model",
        [
            ("primary_email_contact", "emails"),
            ("primary_telephone_contact", "telephones"),
            ("primary_address_contact", "addresses"),
            ("primary_website_contact", "websites"),
            ("primary_banking_contact", "banking"),
        ],
    )
    def test_get_primary_contact(self, mocker: MockerFixture, contact_method, contact_model):
        # Arrange
        entry = Entry()
        mock_contact_property = mocker.patch(
            f"wbcore.contrib.directory.models.Entry.{contact_model}", new_callable=mocker.PropertyMock
        )
        mock_contact_property.return_value.get.return_value.primary = True
        # Act
        method = getattr(entry, contact_method)()
        method()
        # Assert
        mock_contact_property.return_value.get.assert_called_once_with(primary=True)

    @pytest.mark.with_db
    @pytest.mark.django_db
    def test_get_casted_entry(self, test_person, test_company):
        # Arrange
        company_entry = Entry.objects.get(id=test_company.id)
        person_entry = Entry.objects.get(id=test_person.id)
        # Act
        casted_company_entry = company_entry.get_casted_entry()
        casted_company = test_company.get_casted_entry()
        casted_person_entry = person_entry.get_casted_entry()
        # Assert
        assert casted_company_entry == test_company
        assert casted_company == test_company
        assert casted_person_entry == test_person

    def test_delete_additional_fields(self):
        # Arrange
        entry = Entry()
        # Act
        entry.additional_fields["special_key"] = "additional fields special key"
        entry.delete_additional_fields("special_key")
        # Assert
        assert "special_key" not in entry.additional_fields.keys()

    def test_full_name(self, mocker: MockerFixture):
        # Arrange
        first_name = "Foo"
        last_name = "Bar"
        person = Person()
        mocker.patch.object(person, "first_name", first_name)
        mocker.patch.object(person, "last_name", last_name)
        # Assert
        assert person.full_name == f"{last_name} {first_name}"

    @pytest.mark.with_db
    @pytest.mark.django_db
    @pytest.mark.parametrize("with_employer", [True, False])
    def test_str_full(self, with_employer, test_person, test_company, mocker: MockerFixture):
        # Arrange
        name = f"{test_person.first_name} {test_person.last_name}"
        if with_employer:
            company_name = test_company.name
            test_person.employers.add(test_company)
            name += f'{(" (%s)" % company_name)}'
        # Assert
        assert test_person.str_full() == name

    @pytest.mark.with_db
    @pytest.mark.django_db
    def test_soft_deleted_entry_dont_show_in_queryset(self, company_factory, person_factory):
        "ensure the default queryset filter out soft deleted entries"
        person = person_factory.create()
        assert set(Person.objects.all()) == {person}
        person.delete()
        assert not Person.objects.exists()
        company = company_factory.create()
        assert set(Company.objects.all()) == {company}
        company.delete()
        assert not Company.objects.exists()

    @pytest.mark.with_db
    @pytest.mark.django_db
    def test_soft_deleted_person_doesnt_crash_user_registration(self, user_factory, email_contact_factory):
        user = user_factory(profile=None)
        email_contact_factory(address=user.email, entry__is_active=False)
        assert Person.get_or_create_with_user(user)


@pytest.mark.with_db
@pytest.mark.django_db
@pytest.mark.directory_model_tests
class TestUserDeactivation:
    @pytest.fixture
    def test_internal_profile(self):
        return InternalUserFactory().profile

    def test_main_company_removed_from_deactivated_user(self, test_internal_profile):
        # Arrange
        main_company_id = global_preferences_registry.manager()["directory__main_company"]
        main_company = Company.objects.get(id=main_company_id)
        message = handle_user_deactivation(sender=None, instance=test_internal_profile, substitute_profile=None)[0]
        # Assert
        assert message == f"Removed {main_company.computed_str} from {test_internal_profile.computed_str}'s employers"
        assert main_company not in test_internal_profile.employers.all()

    def test_no_substitute_person(self, test_internal_profile):
        # Arrange
        relationship = CMRF(relationship_manager=test_internal_profile)
        # Act
        handle_user_deactivation(sender=None, instance=test_internal_profile, substitute_profile=None)
        # Assert
        assert CMR.objects.get(id=relationship.id).status == CMR.Status.REMOVED

    @pytest.mark.parametrize("exists", [True, False])
    def test_not_approved_substitute_relationships(self, test_internal_profile, test_person, exists):
        # Arrange
        old_relationship = CMRF(relationship_manager=test_internal_profile, status=CMR.Status.PENDINGADD)
        substitute_relationship = (
            CMRF(client=old_relationship.client, relationship_manager=test_person) if exists else None
        )
        # Act
        message = handle_user_deactivation(
            sender=None, instance=test_internal_profile, substitute_profile=test_person
        )[1]
        relationship_id = substitute_relationship.id if exists else old_relationship.id
        relationship_exists = CMR.objects.filter(id=old_relationship.id).exists()
        # Assert
        assert message == f"Assigned 1 manager role(s) to {test_person.computed_str}"
        assert (not relationship_exists) if exists else relationship_exists
        assert CMR.objects.get(id=relationship_id).client.id == old_relationship.client.id
        assert CMR.objects.get(id=relationship_id).relationship_manager.id == test_person.id

    def test_approved_with_substitute_relationships(self, test_internal_profile, test_person):
        # Arrange
        old_relationship = CMRF(relationship_manager=test_internal_profile, primary=True)
        CMRF(relationship_manager=test_internal_profile)
        substitute_relationship = CMRF(
            client=old_relationship.client, relationship_manager=test_person, status=CMR.Status.PENDINGADD
        )
        # Act
        message = handle_user_deactivation(
            sender=None, instance=test_internal_profile, substitute_profile=test_person
        )[1]
        substitute_relationship.refresh_from_db()
        # Assert
        assert message == f"Assigned 2 manager role(s) to {test_person.computed_str}"
        assert substitute_relationship.client.id == old_relationship.client.id
        assert substitute_relationship.relationship_manager.id == test_person.id
        assert substitute_relationship.status == CMR.Status.APPROVED
        assert substitute_relationship.primary is True

    def test_approved_without_substitute_relationships_needs_primary(self, test_internal_profile, test_person):
        # Arrange
        old_relationship = CMRF(relationship_manager=test_internal_profile, primary=True)
        # Act
        message = handle_user_deactivation(
            sender=None, instance=test_internal_profile, substitute_profile=test_person
        )[1]
        relationship_exists = CMR.objects.filter(
            relationship_manager=test_person, client=old_relationship.client, primary=True, status=CMR.Status.APPROVED
        ).exists()
        old_relationship.refresh_from_db()
        # Assert
        assert message == f"Assigned 1 manager role(s) to {test_person.computed_str}"
        assert relationship_exists
        assert old_relationship.status == CMR.Status.REMOVED

    def test_approved_without_substitute_relationships_doesnt_need_primary(self, test_internal_profile, test_person):
        # Arrange
        old_relationship = CMRF(relationship_manager=test_internal_profile)
        CMRF(client=old_relationship.client, primary=True)
        # Act
        message = handle_user_deactivation(
            sender=None, instance=test_internal_profile, substitute_profile=test_person
        )[1]
        relationship_exists = CMR.objects.filter(
            relationship_manager=test_person, client=old_relationship.client, primary=False, status=CMR.Status.APPROVED
        ).exists()
        old_relationship.refresh_from_db()
        # Assert
        assert message == f"Assigned 1 manager role(s) to {test_person.computed_str}"
        assert relationship_exists
        assert old_relationship.status == CMR.Status.REMOVED


@pytest.mark.directory_model_tests
class TestSpecificModelsClientManagerRelationship:
    def _mock_relationship_query(self, is_primary: bool, mocker: MockerFixture, mock_path: str) -> MagicMock:
        mock_filter = mocker.patch(f"{mock_path}.objects.filter")
        mock_exists = mock_filter.return_value.exists
        mock_exists.return_value = is_primary
        mock_update = mock_filter.return_value.update
        mock_update.return_value = None
        return mock_update

    @pytest.fixture
    def test_cmr(self):
        return CMR()

    @pytest.mark.parametrize(
        "method_name, initial_status, expected_status",
        [
            ("submit", CMR.Status.DRAFT, CMR.Status.PENDINGADD),
            ("deny", CMR.Status.PENDINGADD, CMR.Status.DRAFT),
            ("denyremoval", CMR.Status.PENDINGREMOVE, CMR.Status.APPROVED),
            ("approveremoval", CMR.Status.PENDINGREMOVE, CMR.Status.REMOVED),
            ("reinstate", CMR.Status.REMOVED, CMR.Status.PENDINGADD),
        ],
    )
    def test_status_transitions(self, test_cmr, mocker: MockerFixture, method_name, initial_status, expected_status):
        # Arrange
        mocker.patch.object(test_cmr, "status", initial_status)
        mocker.patch("django_fsm.transition", return_value=None)
        # Act
        method = getattr(test_cmr, method_name)
        method()
        # Assert
        assert test_cmr.status == expected_status

    @pytest.mark.parametrize("method_name", ["approve", "mngapprove"])
    @pytest.mark.parametrize("is_primary", [True, False])
    def test_approval_methods(self, test_cmr, method_name, is_primary, mocker: MockerFixture):
        # Arrange
        status = CMR.Status.PENDINGADD if method_name == "approve" else CMR.Status.DRAFT
        mock_path = "wbcore.contrib.directory.models.ClientManagerRelationship"
        mocker.patch(f"{mock_path}.client", new_callable=mocker.PropertyMock)  # Mock client property
        mock_update = self._mock_relationship_query(is_primary, mocker, mock_path)
        mocker.patch("django_fsm.transition", return_value=None)
        mocker.patch.object(test_cmr, "primary", is_primary)
        mocker.patch.object(test_cmr, "status", status)
        # Act
        method = getattr(test_cmr, method_name)
        method()
        # Assert
        assert test_cmr.status == CMR.Status.APPROVED
        assert test_cmr.primary
        if is_primary:
            mock_update.assert_called_once_with(primary=False)
        else:
            mock_update.assert_not_called()

    @pytest.mark.parametrize("is_primary", [True, False])
    def test_make_primary(self, is_primary, test_cmr, mocker: MockerFixture):
        # Arrange
        mocker.patch.object(test_cmr, "primary", is_primary)
        mocker.patch.object(test_cmr, "status", CMR.Status.APPROVED)
        mocker.patch("django_fsm.transition", return_value=None)
        # Act & Assert
        if is_primary:
            with pytest.raises(TransitionNotAllowed):
                test_cmr.makeprimary()
                assert test_cmr.status == CMR.Status.APPROVED
        else:
            test_cmr.makeprimary()
            assert test_cmr.status == CMR.Status.PENDINGADD
        assert test_cmr.primary

    @pytest.mark.parametrize("is_primary", [True, False])
    @pytest.mark.parametrize("exists", [True, False])
    def test_remove(self, test_cmr, is_primary, exists, mocker: MockerFixture):
        # Arrange
        mock_path = "wbcore.contrib.directory.models.ClientManagerRelationship"
        mock_queryset = mocker.MagicMock()
        mock_queryset.filter.return_value.exists.return_value = exists
        mock_exclude = mocker.patch(f"{mock_path}.objects.exclude")
        mock_exclude.return_value = mock_queryset
        mocker.patch(f"{mock_path}.client", new_callable=mocker.PropertyMock)  # Mock client property
        mocker.patch.object(test_cmr, "primary", is_primary)
        mocker.patch.object(test_cmr, "status", CMR.Status.APPROVED)
        mocker.patch("django_fsm.transition", return_value=None)
        # Act & Assert
        if not is_primary and exists:
            test_cmr.remove()
            assert test_cmr.status == CMR.Status.PENDINGREMOVE
        else:
            with pytest.raises(TransitionNotAllowed):
                test_cmr.remove()
                assert test_cmr.status == CMR.Status.APPROVED


@pytest.mark.directory_model_tests
class TestSpecificModelsRelationships:
    def test_relationship_type_str(self, mocker: MockerFixture):
        # Arrange
        title = "Type"
        relationship_type = RelationshipType()
        mocker.patch.object(relationship_type, "title", title)
        # Act & Assert
        assert relationship_type.__str__() == title

    @pytest.mark.with_db
    @pytest.mark.django_db
    def test_relationship_str(self, relationship_factory, relationship_type_factory, person_factory):
        relationship_type = relationship_type_factory(
            title="Type",
            counter_relationship=None,
        )
        from_entry = person_factory(first_name="John", last_name="Doe")
        to_entry = person_factory(first_name="Jane", last_name="Doe")
        rel = relationship_factory(
            relationship_type=relationship_type,
            from_entry=from_entry,
            to_entry=to_entry,
        )
        assert rel.__str__() == "John Doe is Type of Jane Doe"



from django.contrib import admin

from wbcore.contrib.workflow.models import Transition


@admin.register(Transition)
class TransitionAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "from_step", "to_step", "icon")



from typing import TYPE_CHECKING

from django.contrib.messages import info
from django.db.models import BooleanField, Case, F, Q, Value, When
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _

from wbcore import viewsets
from wbcore.contrib.directory.models import (
    AddressContact,
    BankingContact,
    Company,
    EmailContact,
    Entry,
    Person,
    SocialMediaContact,
    TelephoneContact,
    WebsiteContact,
)
from wbcore.contrib.directory.models import EmployerEmployeeRelationship as EER

from ..filters import (
    AddressContactCompanyFilter,
    AddressContactPersonFilter,
    BankingContactCompanyFilter,
    BankingContactFilter,
    BankingContactPersonFilter,
    EmailContactCompanyFilter,
    EmailContactPersonFilter,
    SocialMediaContactCompanyFilter,
    SocialMediaContactPersonFilter,
    TelephoneContactCompanyFilter,
    TelephoneContactPersonFilter,
    WebsiteContactCompanyFilter,
    WebsiteContactPersonFilter,
)
from ..serializers import (
    AddressContactSerializer,
    BankingContactRepresentationSerializer,
    BankingContactSerializer,
    EmailContactRepresentationSerializer,
    EmailContactSerializer,
    ReadOnlyBankingContactSerializer,
    SocialMediaContactRepresentationSerializer,
    SocialMediaContactSerializer,
    TelephoneContactRepresentationSerializer,
    TelephoneContactSerializer,
    WebsiteContactRepresentationSerializer,
    WebsiteContactSerializer,
)
from .buttons import TelephoneContactButtonConfig
from .display.contacts import (
    AddressContactDisplay,
    AddressContactEntryDisplay,
    BankingContactDisplay,
    BankingContactEntryDisplay,
    EmailContactDisplay,
    EmailContactEntryDisplay,
    SocialMediaContactEntryDisplay,
    TelephoneContactDisplay,
    TelephoneContactEntryDisplay,
    WebsiteContactEntryDisplay,
)
from .endpoints.contacts import (
    AddressContactEntryEndpointConfig,
    BankingContactEndpointConfig,
    BankingContactEntryEndpointConfig,
    EmailContactEntryEndpointConfig,
    SocialMediaContactEntryEndpointConfig,
    TelephoneContactEntryEndpointConfig,
    WebsiteContactEntryEndpointConfig,
)
from .previews.contacts import AddressPreviewConfig
from .titles.contacts import (
    AddressContactEntryTitleConfig,
    AddressContactTitleConfig,
    BankingContactEntryTitleConfig,
    BankingContactTitleConfig,
    EmailContactEntryTitleConfig,
    EmailContactTitleConfig,
    SocialMediaContactEntryTitleConfig,
    TelephoneContactEntryTitleConfig,
    TelephoneContactTitleConfig,
    WebsiteContactEntryTitleConfig,
)

if TYPE_CHECKING:
    _Base = viewsets.ModelViewSet
else:
    _Base = object


def get_propagated_contact(qs, entry_id, propage=True):
    """
    Convenience function.
    Annotates if the contact is from a Company or Person.
    """

    entry = get_object_or_404(Entry, pk=entry_id)
    if not entry.is_company:
        person = Person.objects.get(id=entry.id)
        if person.employers.exists():
            return qs.filter(Q(entry__id=person.id) | Q(entry__in=person.employers.all())).annotate(
                company_contact=Case(
                    When(entry__entry_type=Company.__name__, then=Value(True)),
                    default=Value(False),
                    output_field=BooleanField(),
                )
            )
    return qs.filter(entry__id=entry.id)


class ContactModelMixin(_Base):
    def add_messages(self, request, instance=None, **kwargs):
        if instance and instance.primary:
            contact = self.queryset.model._meta.verbose_name
            info(
                request,
                _(
                    "This is the primary {contact}. In order to delete the primary {contact}"
                    + " you have to specify a different primary contact or make it the only contact. "
                    + "To change primary status make a different contact primary instead."
                ).format(contact=contact),
            )

    def get_queryset(self):
        if "entry_id" in self.kwargs:
            entry: Entry = Entry.all_objects.get(id=self.kwargs["entry_id"])
            if not entry.is_company:
                person = entry.get_casted_entry()
                if (
                    EER.objects.filter(employee=person, primary=True).exists()
                    and EER.objects.filter(employee=person).count() > 1
                ):
                    return (
                        super()
                        .get_queryset()
                        .exclude(
                            entry__in=EER.objects.filter(employee=person, primary=False).values_list(
                                "employer", flat=True
                            )
                        )
                    )
        return super().get_queryset()


class EmailContactRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ("address",)
    ordering = ["address", "pk"]
    ordering_fields = ["address"]
    serializer_class = EmailContactRepresentationSerializer
    queryset = EmailContact.objects.all().select_related("entry")

    filterset_fields = {
        "entry": ["exact"],
        "location": ["exact"],
        "primary": ["exact"],
    }


class EmailContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/email.md"
    title_config_class = EmailContactTitleConfig
    display_config_class = EmailContactDisplay
    ordering = ["primary", "location", "pk"]
    search_fields = ("address",)
    ordering_fields = ("address", "location")
    queryset = EmailContact.objects.select_related("entry")
    serializer_class = EmailContactSerializer


class AddressContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/address.md"
    title_config_class = AddressContactTitleConfig
    display_config_class = AddressContactDisplay
    filterset_fields = {
        "location": ["exact"],
        "street": ["exact", "icontains"],
        "street_additional": ["exact", "icontains"],
        "zip": ["exact", "icontains"],
    }
    preview_config_class = AddressPreviewConfig

    search_fields = ("street", "zip", "geography_city__name")
    ordering = ["primary", "location", "pk"]
    ordering_fields = [
        "zip",
        "geography_country__name",
        "street",
        "location",
        "street_additional",
        "geography_city__name",
        "geography_state__name",
        "geography_continent__name",
    ]
    queryset = AddressContact.objects.select_related("entry", "geography_city")
    serializer_class = AddressContactSerializer


class BankingContactRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ("institute", "iban", "swift_bic", "entry__computed_str")
    ordering_fields = ["status"]
    ordering = ["status", "pk"]
    serializer_class = BankingContactRepresentationSerializer
    queryset = BankingContact.objects.all()


class BankingContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/banking.md"
    filterset_class = BankingContactFilter
    search_fields = ("iban", "institute", "swift_bic")
    ordering = ["status", "primary", "location", "pk"]
    ordering_fields = [
        "additional_information",
        "currency__title",
        "entry__computed_str",
        "iban",
        "institute",
        "institute_additional",
        "location",
        "status",
        "swift_bic",
        "edited",
    ]
    endpoint_config_class = BankingContactEndpointConfig
    queryset = BankingContact.objects.select_related("currency", "entry")
    serializer_class = BankingContactSerializer
    display_config_class = BankingContactDisplay
    title_config_class = BankingContactTitleConfig

    def get_serializer_class(self):
        if self.new_mode or (
            "pk" in self.kwargs and (contact := self.get_object()) and contact.status == BankingContact.Status.DRAFT
        ):
            return BankingContactSerializer
        return ReadOnlyBankingContactSerializer


class EmailContactEntryViewSet(EmailContactViewSet):
    display_config_class = EmailContactEntryDisplay
    title_config_class = EmailContactEntryTitleConfig
    endpoint_config_class = EmailContactEntryEndpointConfig
    serializer_class = EmailContactSerializer

    def get_queryset(self):
        return get_propagated_contact(super().get_queryset(), self.kwargs["entry_id"])

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return EmailContactCompanyFilter
        return EmailContactPersonFilter


class AddressContactEntryViewSet(AddressContactViewSet):
    display_config_class = AddressContactEntryDisplay
    title_config_class = AddressContactEntryTitleConfig
    endpoint_config_class = AddressContactEntryEndpointConfig

    def get_queryset(self):
        qs = get_propagated_contact(super().get_queryset(), self.kwargs["entry_id"])
        qs = qs.annotate(
            geography_state=F("geography_city__parent__name"),
            geography_country=F("geography_city__parent__parent__name"),
            geography_continent=F("geography_city__parent__parent__parent__name"),
        )
        return qs

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return AddressContactCompanyFilter
        return AddressContactPersonFilter


class TelephoneContactRepresentationViewSet(viewsets.RepresentationViewSet):
    search_fields = ("number",)
    ordering = ["primary", "pk"]
    ordering_fields = ["primary"]
    serializer_class = TelephoneContactRepresentationSerializer
    queryset = TelephoneContact.objects.all()


class TelephoneContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/telephone.md"
    ordering = ["primary", "location", "pk"]
    search_fields = ["number"]
    ordering_fields = ["number", "telephone_type", "location"]
    filterset_fields = {
        "primary": ["exact"],
        "location": ["exact"],
        "entry": ["exact"],
        "number": ["contains"],
        "telephone_type": ["exact"],
    }
    queryset = TelephoneContact.objects.select_related("entry")
    serializer_class = TelephoneContactSerializer
    display_config_class = TelephoneContactDisplay
    title_config_class = TelephoneContactTitleConfig
    button_config_class = TelephoneContactButtonConfig


class TelephoneContactEntryViewSet(TelephoneContactViewSet):
    display_config_class = TelephoneContactEntryDisplay
    title_config_class = TelephoneContactEntryTitleConfig
    endpoint_config_class = TelephoneContactEntryEndpointConfig
    search_fields = ["number"]

    def get_queryset(self):
        return get_propagated_contact(super().get_queryset(), self.kwargs["entry_id"])

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return TelephoneContactCompanyFilter
        return TelephoneContactPersonFilter


class WebsiteContactRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = WebsiteContactRepresentationSerializer
    queryset = WebsiteContact.objects.all()


class WebsiteContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    serializer_class = WebsiteContactSerializer
    queryset = WebsiteContact.objects.select_related("entry")


class WebsiteContactEntryViewSet(ContactModelMixin, viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/website.md"
    ordering = ["primary", "location", "pk"]
    ordering_fields = ["url", "location"]
    search_fields = ["url"]
    queryset = WebsiteContact.objects.select_related("entry")
    serializer_class = WebsiteContactSerializer
    display_config_class = WebsiteContactEntryDisplay
    title_config_class = WebsiteContactEntryTitleConfig
    endpoint_config_class = WebsiteContactEntryEndpointConfig

    def get_queryset(self):
        return get_propagated_contact(super().get_queryset(), self.kwargs["entry_id"])

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return WebsiteContactCompanyFilter
        return WebsiteContactPersonFilter


class BankingContactEntryViewSet(BankingContactViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/bankingentry.md"
    display_config_class = BankingContactEntryDisplay
    title_config_class = BankingContactEntryTitleConfig
    endpoint_config_class = BankingContactEntryEndpointConfig

    def get_queryset(self):
        return get_propagated_contact(super().get_queryset(), entry_id=self.kwargs["entry_id"])

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return BankingContactCompanyFilter
        return BankingContactPersonFilter


class SocialMediaContactViewSet(ContactModelMixin, viewsets.ModelViewSet):
    search_fields = ("platform",)
    ordering_fields = ("location", "platform", "url")
    ordering = ("platform", "location", "pk")
    queryset = SocialMediaContact.objects.select_related("entry")
    serializer_class = SocialMediaContactSerializer


class SocialMediaContactEntryViewSet(SocialMediaContactViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/socialmedia.md"
    display_config_class = SocialMediaContactEntryDisplay
    endpoint_config_class = SocialMediaContactEntryEndpointConfig
    title_config_class = SocialMediaContactEntryTitleConfig

    def get_queryset(self):
        return get_propagated_contact(super().get_queryset(), self.kwargs["entry_id"])

    def get_filterset_class(self, request):
        entry = get_object_or_404(Entry, id=self.kwargs["entry_id"])
        if entry.is_company:
            return SocialMediaContactCompanyFilter
        return SocialMediaContactPersonFilter


class SocialMediaContactRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = SocialMediaContactRepresentationSerializer
    queryset = SocialMediaContact.objects.all()



from .condition import ConditionTitleConfig
from .data import DataTitleConfig
from .process import ProcessStepTitleConfig, ProcessTitleConfig, AssignedProcessStepTitleConfig
from .step import (
    DecisionStepTitleConfig,
    EmailStepTitleConfig,
    FinishStepTitleConfig,
    JoinStepTitleConfig,
    ScriptStepTitleConfig,
    SplitStepTitleConfig,
    StepTitleConfig,
    UserStepTitleConfig,
    StartStepTitleConfig,
)
from .transition import TransitionTitleConfig
from .workflow import WorkflowTitleConfig



import pytest
from django.utils import timezone
from pytest_mock import MockerFixture

from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.currency.serializers import (
    CurrencyFXRatesModelSerializer,
    CurrencyModelSerializer,
    CurrencyRepresentationSerializer,
)


@pytest.mark.serializer_tests
class TestSerializers:
    @pytest.fixture
    def currency(self, mocker: MockerFixture):
        currency = Currency()
        mocker.patch.object(currency, "id", 1)
        mocker.patch.object(currency, "key", "abc")
        mocker.patch.object(currency, "title", "ABC")
        mocker.patch.object(currency, "symbol", "$")
        return currency

    @pytest.fixture
    def currency_fx_rate(self, mocker: MockerFixture, currency):
        currency_fx_rate = CurrencyFXRates()
        mocker.patch.object(currency_fx_rate, "id", 1)
        mocker.patch.object(currency_fx_rate, "date", timezone.now().date().isoformat())
        mocker.patch.object(currency_fx_rate, "value", "1.150000")
        mocker.patch.object(currency_fx_rate, "currency_id", currency.id)
        return currency_fx_rate

    @pytest.fixture
    def expected_data(self, currency):
        return {"id": currency.id, "title": currency.title, "key": currency.key, "symbol": currency.symbol}

    @pytest.fixture
    def expected_data_fx_rate(self, currency_fx_rate):
        return {
            "id": currency_fx_rate.id,
            "value": currency_fx_rate.value,
            "date": currency_fx_rate.date,
            "currency": currency_fx_rate.currency_id,
        }

    def test_currency_representation_serializer(self, mocker: MockerFixture, currency, expected_data):
        detail = f"/wbcore/currency/currency/{currency.id}/"
        mocker.patch("wbcore.serializers.fields.fields.HyperlinkField", return_value=detail)
        serializer = CurrencyRepresentationSerializer(instance=currency)
        expected_data["_detail"] = detail
        assert serializer.data == expected_data

    def test_currency_model_serializer(self, mocker: MockerFixture, currency, expected_data):
        mocker.patch(
            "wbcore.serializers.fields.list.SparklineField.to_representation", return_value={"Rates": [1, 2, 3]}
        )
        serializer = CurrencyModelSerializer(instance=currency)
        expected_data.update({"rates_sparkline": {"Rates": [1, 2, 3]}, "_additional_resources": {}})
        assert serializer.data == expected_data

    def test_currency_fx_rate_model_serializer(self, currency_fx_rate, expected_data_fx_rate):
        serializer = CurrencyFXRatesModelSerializer(instance=currency_fx_rate)
        assert serializer.data == expected_data_fx_rate



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

TEAM_FIELDS = [
    dp.Field(key="city", label=_("City")),
    dp.Field(key="founded_date", label=_("Founded Date")),
    dp.Field(key="coach", label=_("Coach"), tooltip=dp.Tooltip(key="coach_tooltip")),
    dp.Field(key="home_stadium", label=_("Home Stadium")),
    dp.Field(
        key="computed_str",
        label=_("Name"),
        formatting_rules=[dp.FormattingRule(style={"fontWeight": "bold"}, condition=("==", "Altona 93"))],
    ),
]


class TeamDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=TEAM_FIELDS,
            tree=True,
            tree_group_field="computed_str",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_depth=1,
                    lookup="current_players__id",
                    filter_key="current_team",
                    list_endpoint=reverse("example_app:treeviewplayer-list", args=[], request=self.request),
                    reorder_endpoint=reverse("example_app:treeviewplayer-list", args=[], request=self.request)
                    + "{{id}}/reorder/",
                    reparent_endpoint=reverse("example_app:treeviewplayer-list", args=[], request=self.request)
                    + "{{id}}/reparent/",
                    parent_field="id",
                )
            ],
        )

    def get_instance_display(self) -> Display:
        players_section = Section(
            key="players_section",
            title=_("Players"),
            display=Display(
                pages=[
                    Page(
                        title=_("Players"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["players_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="players_inline", endpoint="players")],
                            )
                        },
                    ),
                ]
            ),
        )
        matches_section = Section(
            key="matches_section",
            collapsible=False,
            title=_("Recent Matches"),
            display=Display(
                pages=[
                    Page(
                        title=_("Recent Matches"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["matches_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="matches_inline", endpoint="recent_matches")],
                            )
                        },
                    ),
                ]
            ),
        )

        is_create_display = "pk" not in self.view.kwargs

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=(
                                [
                                    [
                                        "name",
                                        "coach",
                                        "matches_section",
                                        "matches_section",
                                    ],
                                    [
                                        "home_stadium",
                                        ".",
                                        "matches_section",
                                        "matches_section",
                                    ],
                                    [".", ".", "matches_section", "matches_section"],
                                    [
                                        "players_section",
                                        "players_section",
                                        "players_section",
                                        "players_section",
                                    ],
                                ]
                                if not is_create_display
                                else [
                                    ["name", "name", "coach", "coach"],
                                    ["home_stadium", "home_stadium", ".", "."],
                                ]
                            ),
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                            sections=[players_section, matches_section],
                        ),
                        lte(1000): Layout(
                            grid_template_areas=(
                                [
                                    ["name", "coach"],
                                    ["home_stadium", "."],
                                    ["matches_section", "matches_section"],
                                    ["players_section", "players_section"],
                                ]
                                if not is_create_display
                                else [["name", "coach"], ["home_stadium", "."]]
                            ),
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[players_section, matches_section],
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["coach"],
                                ["home_stadium"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Details"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["city", "founded_date"],
                                ["phone_number", "email"],
                                ["duration_since_last_win", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["city"],
                                ["founded_date"],
                                ["phone_number"],
                                ["email"],
                                ["duration_since_last_win"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ],
        )


class TeamStadiumDisplayConfig(TeamDisplayConfig):
    def get_list_display(self) -> ListDisplay:
        fields = TEAM_FIELDS.copy()
        fields.insert(0, dp.Field(key="computed_str", label=_("Name")))
        fields.pop(4)
        return dp.ListDisplay(fields=fields)

    def get_instance_display(self) -> Display:
        if "pk" in self.view.kwargs:
            return super().get_instance_display()

        return Display(
            pages=[
                Page(
                    title=_("Main Information"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["name", "coach", "league"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(1000): Layout(
                            grid_template_areas=[["name", "coach"], ["league", "."]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                        lte(450): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["coach"],
                                ["league"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
                Page(
                    title=_("Details"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["city", "founded_date"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[["city"], ["founded_date"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ],
        )



from datetime import timedelta

import pytz
from psycopg.types.range import DateRange, TimestamptzRange
from rest_framework import serializers
from timezone_field.choices import standard, with_gmt_offset
from timezone_field.rest_framework import TimeZoneSerializerField

from .mixins import RangeMixin, WBCoreSerializerFieldMixin
from .number import NumberFieldMixin
from .types import WBCoreType


class DateTimeField(WBCoreSerializerFieldMixin, serializers.DateTimeField):
    field_type = WBCoreType.DATETIME.value


class DateField(WBCoreSerializerFieldMixin, serializers.DateField):
    field_type = WBCoreType.DATE.value


class TimeField(WBCoreSerializerFieldMixin, serializers.TimeField):
    field_type = WBCoreType.TIME.value


class ShortcutMixin(WBCoreSerializerFieldMixin):
    def __init__(self, shortcuts: list | None = None, *args, **kwargs):
        self.shortcuts = shortcuts
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        if self.shortcuts is not None:
            representation["shortcuts"] = self.shortcuts
        return key, representation


class DateRangeField(RangeMixin, ShortcutMixin, serializers.DateField):
    field_type = WBCoreType.DATERANGE.value
    internal_field = DateRange

    def __init__(
        self,
        *args,
        outward_bounds_transform="[)",
        inward_bounds_transform="[)",
        **kwargs,
    ):
        self.outward_bounds_transform = outward_bounds_transform  # Allow to specify another bound representation than the default and used for db (]. If specified, will switch the range around to and from the serializer
        self.inward_bounds_transform = inward_bounds_transform
        if self.outward_bounds_transform not in ["[]", "[)", "()", "(]"]:
            raise ValueError(f"Outward bound transform {self.outward_bounds_transform} is not a valid choice")
        if self.inward_bounds_transform not in ["[]", "[)", "()", "(]"]:
            raise ValueError(f"Inward bound transform {self.inward_bounds_transform} is not a valid choice")

        super().__init__(*args, **kwargs)

    def _transform_range(self, lower, upper, inward=False):
        """
        Private utility function that shift the range by one operator
        Args:
            lower: Lower bound
            upper: Upper bound
            inverse: If True, the transform is back to its original bound transform

        Returns:
            The shifted bound
        """
        bound_transform = self.inward_bounds_transform if inward else self.outward_bounds_transform
        if lower and bound_transform[0] == "(":
            lower = lower + timedelta(days=1) if inward else lower - timedelta(days=1)
        if upper and bound_transform[1] == "]":
            upper = upper + timedelta(days=1) if inward else upper - timedelta(days=1)
        return lower, upper


class DateTimeRangeField(RangeMixin, ShortcutMixin, serializers.DateTimeField):
    field_type = WBCoreType.DATETIMERANGE.value
    internal_field = TimestamptzRange

    def __init__(self, *args, lower_time_choices=None, upper_time_choices=None, **kwargs):
        self.lower_time_choices = lower_time_choices
        if (
            self.lower_time_choices
            and not isinstance(self.lower_time_choices, list)
            and not callable(self.lower_time_choices)
        ):
            raise ValueError("lower_time_choices can only be a static list or a callable.")
        self.upper_time_choices = upper_time_choices
        if (
            self.upper_time_choices
            and not isinstance(self.upper_time_choices, list)
            and not callable(self.upper_time_choices)
        ):
            raise ValueError("upper_time_choices can only be a static list or a callable.")
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        if self.lower_time_choices is not None:
            if callable(self.lower_time_choices):
                representation["lower_time_choices"] = self.lower_time_choices(self, request)
            else:
                representation["lower_time_choices"] = self.lower_time_choices

        if self.upper_time_choices is not None:
            if callable(self.upper_time_choices):
                representation["upper_time_choices"] = self.upper_time_choices(self, request)
            else:
                representation["upper_time_choices"] = self.upper_time_choices
        if timezone := getattr(self, "timezone", None):
            representation["timezone"] = str(timezone)
        return key, representation


class DurationField(NumberFieldMixin, WBCoreSerializerFieldMixin, serializers.DurationField):
    field_type = WBCoreType.DURATION.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.precision = 0


class TimeZoneField(WBCoreSerializerFieldMixin, TimeZoneSerializerField):
    field_type = WBCoreType.SELECT.value

    def __init__(self, choices=None, choices_display=None, *args, **kwargs):
        if choices:
            values, displays = zip(*choices)
        else:
            values = pytz.common_timezones
            displays = None

        if choices_display == "WITH_GMT_OFFSET":
            choices = with_gmt_offset(values, use_pytz=self.use_pytz)
        elif choices_display == "STANDARD":
            choices = standard(values)
        elif choices_display is None:
            choices = zip(values, displays) if displays else standard(values)
        else:
            raise ValueError(f"Unrecognized value for kwarg 'choices_display' of '{choices_display}'")

        self.choices = choices
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["choices"] = [{"value": k, "label": v} for k, v in self.choices]
        return key, representation



import pycountry
from django.forms import ValidationError
from django.utils.translation import gettext as _
from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import IntegerPreference, StringPreference

from .models.entries import Company

general = Section("directory")


@global_preferences_registry.register
class MainCompanyPreference(IntegerPreference):
    section = general
    name = "main_company"
    default = 0

    verbose_name = _("Main Workbench Company")
    help_text = _(
        'The main company of the Workbench. All employees of this company are considered to have "employee" status and have more permissions than other users.'
    )

    def validate(self, value):
        if not Company.objects.filter(id=value).exists():
            raise ValidationError(_("Please select a valid company."))


@global_preferences_registry.register
class CountryCodePreference(StringPreference):
    section = general
    name = "main_country_code"
    default = "US"
    verbose_name = _("Main Country Code")
    help_text = _(
        "Determines the country code currently used exclusively for parsing of phone numbers. Should be the country where the phone number is being dialled from."
    )

    def validate(self, value):
        if not pycountry.countries.get(alpha_2=value):
            raise ValidationError(_("Only exact alpha_2 country codes (e.g. DE, CH) allowed."))


@global_preferences_registry.register
class PersonSalutationStringPreference(StringPreference):
    section = general
    name = "person_salutation"
    default = _("Dear {0} {1}")

    verbose_name = _("Person Salutation")
    help_text = _("Salutation that can be used in mails for a person")


@global_preferences_registry.register
class CompanySalutationStringPreference(StringPreference):
    section = general
    name = "company_salutation"
    default = _("To everyone from {0}")

    verbose_name = _("Company Salutation")
    help_text = _("Salutation that can be used in mails for a company")



# Generated by Django 5.0.10 on 2025-01-31 09:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0006_notification_created'),
    ]

    operations = [
        migrations.AddField(
            model_name='notificationtype',
            name='resource_button_label',
            field=models.CharField(default='Open Resource', max_length=128),
            preserve_default=False,
        ),
    ]



from django.conf import settings
from django.contrib import admin
from django_better_admin_arrayfield.models.fields import ArrayField
from wbcore.contrib.color.forms import DynamicArrayColorWidget
from wbcore.contrib.color.models import ColorGradient


@admin.register(ColorGradient)
class ColorGradientModelAdmin(admin.ModelAdmin):
    class Media:
        if settings.DEBUG:
            js = [
                "colorfield/jscolor/jscolor.js",
                "colorfield/colorfield.js",
                "js/django_better_admin_arrayfield.min.js",
            ]
        else:
            js = [
                "colorfield/jscolor/jscolor.min.js",
                "colorfield/colorfield.js",
                "js/django_better_admin_arrayfield.min.js",
            ]
        css = {"all": ("css/django_better_admin_arrayfield.min.css",)}

    formfield_overrides = {
        ArrayField: {"widget": DynamicArrayColorWidget},
    }
    list_display = ["title"]



import json
from datetime import datetime
from uuid import UUID

from django.db.models import Q
from django.http.request import HttpRequest
from django.http.response import FileResponse, HttpResponse
from django.shortcuts import get_object_or_404
from django.views.generic import View
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response

from wbcore import viewsets
from wbcore.contrib.documents.filters import (
    ShareableLinkAccessFilter,
    ShareableLinkFilter,
)
from wbcore.contrib.documents.models import ShareableLink, ShareableLinkAccess
from wbcore.contrib.documents.serializers import (
    ShareableLinkAccessModelSerializer,
    ShareableLinkAccessRepresentationSerializer,
    ShareableLinkModelSerializer,
    ShareableLinkRepresentationSerializer,
)
from wbcore.contrib.documents.viewsets.buttons import ShareableLinkModelButtonConfig
from wbcore.contrib.documents.viewsets.display import (
    ShareableLinkAccessModelDisplay,
    ShareableLinkModelDisplay,
)
from wbcore.contrib.documents.viewsets.endpoints import (
    ShareableLinkAccessEndpointConfig,
    ShareableLinkEndpointConfig,
)
from wbcore.contrib.documents.viewsets.titles import ShareableLinkModelTitleConfig


class DownloadShareableLinkView(View):
    def get(self, request: HttpRequest, uuid: UUID) -> HttpResponse:
        link = get_object_or_404(ShareableLink, uuid=uuid)
        if link.is_valid() and link.document.file.name:
            json_meta = json.loads(json.dumps(request.META, default=str))
            link.access(metadata=json_meta)
            if (content_type := link.document.content_type) and (filename := link.document.filename):
                with link.document.file.open("r") as f:
                    return FileResponse(f, as_attachment=True, filename=filename, content_type=content_type)
        return HttpResponse("Link invalid", status=status.HTTP_204_NO_CONTENT)


class ShareableLinkModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "documents/markdown/documentation/shareablelink.md"
    search_fields = ("uuid",)
    ordering = ["valid_until"]
    ordering_fields = (
        "valid_until",
        "one_time_link",
    )
    queryset = ShareableLink.objects.all()
    serializer_class = ShareableLinkModelSerializer
    title_config_class = ShareableLinkModelTitleConfig
    filterset_class = ShareableLinkFilter
    endpoint_config_class = ShareableLinkEndpointConfig
    button_config_class = ShareableLinkModelButtonConfig
    display_config_class = ShareableLinkModelDisplay

    @action(detail=True, methods=["PATCH"])
    def invalidate(self, request, pk=None):
        shareable_link = ShareableLink.objects.get(pk=pk)
        if shareable_link.manual_invalid:
            return Response({}, status=status.HTTP_400_BAD_REQUEST)
        shareable_link.manual_invalid = True
        shareable_link.save()
        return Response({}, status=status.HTTP_200_OK)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                valid=Q(manual_invalid=False) & (Q(valid_until__isnull=True) | Q(valid_until__gte=datetime.now()))
            )
        )


class ShareableLinkRepresentationViewSet(viewsets.RepresentationViewSet):
    serializer_class = ShareableLinkRepresentationSerializer
    queryset = ShareableLink.objects.all()
    search_fields = ("document",)


class ShareableLinkAccessModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "documents/markdown/documentation/shareablelinkaccess.md"
    ordering = ["accessed"]
    ordering_fields = (
        "ipaddress",
        "metadata",
        "accessed",
    )
    queryset = ShareableLinkAccess.objects.all()
    serializer_class = ShareableLinkAccessModelSerializer
    endpoint_config_class = ShareableLinkAccessEndpointConfig
    filterset_class = ShareableLinkAccessFilter

    display_config_class = ShareableLinkAccessModelDisplay


class ShareableLinkAccessRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = ShareableLinkAccess.objects.all()
    serializer_class = ShareableLinkAccessRepresentationSerializer
    search_fields = ("ipaddress",)



# Generated by Django 4.1.8 on 2023-04-26 11:40

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0003_calendaritem_endpoint_basename"),
    ]

    operations = [
        migrations.AlterField(
            model_name="calendaritem",
            name="item_type",
            field=models.CharField(default="agenda.CalendarItem", max_length=255, verbose_name="Type"),
        ),
    ]



from .condition import ConditionAdmin
from .data import DataAdmin, DataValueAdmin
from .display import DisplayAdmin
from .process import ProcessAdmin
from .step import (
    DecisionStepAdmin,
    EmailStepAdmin,
    FinishStepAdmin,
    JoinStepAdmin,
    SplitStepAdmin,
    UserStepAdmin,
    StartStepAdmin,
)
from .transition import TransitionAdmin
from .workflow import WorkflowAdmin



# Generated by Django 4.2.6 on 2023-11-16 10:09

import colorfield.fields
import django_better_admin_arrayfield.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [("wbcore", "0013_delete_colorgradient")]

    state_operations = [
        migrations.CreateModel(
            name="ColorGradient",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(default="", max_length=128)),
                (
                    "colors",
                    django_better_admin_arrayfield.models.fields.ArrayField(
                        base_field=colorfield.fields.ColorField(
                            default="#FF0000", image_field=None, max_length=18, samples=None
                        ),
                        default=list,
                        size=None,
                    ),
                ),
            ],
        ),
    ]

    operations = state_operations



import pytest
from django.forms.models import model_to_dict
from rest_framework.test import APITestCase
from wbcore.contrib.example_app.factories import RoleFactory
from wbcore.contrib.example_app.serializers import RoleModelSerializer


@pytest.mark.django_db
class TestRoleModelSerializer(APITestCase):
    def test_role_serializer(self):
        role_data: dict = model_to_dict(RoleFactory.build())
        role_serializer = RoleModelSerializer(data=role_data)
        self.assertTrue(role_serializer.is_valid())



import pytest
from django.db import models
from faker import Faker

from wbcore.utils.models import PrimaryMixin

fake = Faker()


class TestPrimaryModel(PrimaryMixin):
    PRIMARY_ATTR_FIELDS = ["pivot"]

    pivot = models.CharField(max_length=32, blank=True, null=True)
    field = models.CharField(max_length=32)


@pytest.mark.django_db
class TestPrimaryMixin:
    @pytest.fixture()
    def primary_instance(self):
        return TestPrimaryModel.objects.create(primary=True, pivot="main", field=fake.word())

    def test_get_related_queryset(self):
        instance1 = TestPrimaryModel.objects.create(primary=False, pivot="main", field=fake.word())
        instance2 = TestPrimaryModel.objects.create(primary=False, pivot="main", field=fake.word())
        instance_other = TestPrimaryModel.objects.create(primary=False, pivot="other", field=fake.word())
        assert set(instance1.get_related_queryset()) == {
            instance1,
            instance2,
        }  # Test the queryset are all instance related to pivot "main"
        assert set(instance2.get_related_queryset()) == {
            instance1,
            instance2,
        }  # Test the queryset are all instance related to pivot "main"
        assert set(instance_other.get_related_queryset()) == {
            instance_other
        }  # Test the queryset are all instance related to pivot "other"

    @pytest.mark.parametrize("field", ("Foo Bar"))
    def test_saving_primary(self, field):
        initial_instance = TestPrimaryModel.objects.create(primary=False, pivot="main", field=field)
        assert (
            initial_instance.primary is True
        )  # assert that a non primary saved instance become primary in absence of already primary instance for that pivot
        second_instance = TestPrimaryModel.objects.create(primary=True, pivot="main", field=field)
        assert (
            second_instance.primary is True
        )  # assert that a primary saved instance become primary and the previous primary is unassigned
        initial_instance.refresh_from_db()
        assert initial_instance.primary is False

        unlinked_instance = TestPrimaryModel.objects.create(primary=False, pivot=None, field=field)
        assert unlinked_instance.primary is False  # check that primary logic doesn't apply for unlink instance

    def test_simple_deletion(self, primary_instance):
        assert primary_instance.pivot is not None
        primary_instance.delete()
        primary_instance.refresh_from_db()
        assert primary_instance.pivot is None

    def test_real_deletion(self, primary_instance):
        assert primary_instance.pivot is not None
        primary_instance.delete(no_deletion=False)
        with pytest.raises(TestPrimaryModel.DoesNotExist):
            primary_instance.refresh_from_db()

    def test_deletion_forward_primary_to_next_related_objects(self, primary_instance):
        second_instance = TestPrimaryModel.objects.create(
            primary=False, pivot=primary_instance.pivot, field=fake.word()
        )
        assert second_instance.primary is False  # Basic check
        assert primary_instance.primary is True  # Basic check
        primary_instance.delete()

        second_instance.refresh_from_db()
        primary_instance.refresh_from_db()

        assert primary_instance.pivot is None  # Check if deleted primary is simply unliked
        assert primary_instance.primary is False  # and set to false
        assert second_instance.primary is True  # Check if next non primary related objects become primary



from django.db.models import Case, Exists, OuterRef, Q, Value, When
from rest_framework.decorators import action
from rest_framework.response import Response

from wbcore import viewsets

from ..contrib.icons import WBIcon
from .buttons import ReleaseNotesButtonConfig
from .display import ReleaseNoteDisplayConfig
from .filters import ReleaseNoteFilterSet
from .models import ReleaseNote
from .serializers import ReleaseNoteModelSerializer


class ReleaseNoteReadOnlyModelViewSet(viewsets.ReadOnlyModelViewSet):
    display_config_class = ReleaseNoteDisplayConfig
    button_config_class = ReleaseNotesButtonConfig

    queryset = ReleaseNote.objects.all()
    serializer_class = ReleaseNoteModelSerializer
    filterset_class = ReleaseNoteFilterSet
    search_fields = ("module", "version", "summary")
    ordering_fields = ("module", "version", "release_date")
    ordering = ("-release_date", "-version")

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                user_read=Exists(ReleaseNote.objects.filter(id=OuterRef("id"), read_by=self.request.user)),
                user_read_icon=Case(
                    When(user_read=True, then=Value(WBIcon.VIEW.icon)), default=Value(WBIcon.IGNORE.icon)
                ),
            )
        )

    def retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        if not obj.read_by.filter(id=self.request.user.id).exists():
            obj.read_by.add(self.request.user)
        return super().retrieve(request, *args, **kwargs)

    @action(methods=["GET"], detail=False)
    def unread_count(self, request, pk=None):
        return Response({"count": ReleaseNote.objects.filter(~Q(read_by=self.request.user)).count()})

    @action(methods=["PATCH"], detail=False)
    def mark_all_as_read(self, request, pk=None):
        for rl in ReleaseNote.objects.filter(~Q(read_by=self.request.user)):
            rl.read_by.add(self.request.user)
        return Response({})



from .condition import ConditionDisplayConfig
from .data import DataDisplayConfig
from .process import ProcessDisplayConfig, ProcessStepDisplayConfig, AssignedProcessStepDisplayConfig
from .step import (
    BaseStepDisplayConfig,
    EmailStepDisplayConfig,
    FinishDisplayConfig,
    JoinStepDisplayConfig,
    ScriptStepDisplayConfig,
    StepDisplayConfig,
    UserStepDisplayConfig,
    StartStepDisplay,
)
from .transition import TransitionDisplayConfig
from .workflow import WorkflowDisplayConfig



from typing import Optional

from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.directory.models import BankingContact, Entry
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

CONTACT_PERSON_LEGEND = [
    dp.Legend(
        key="primary",
        items=[dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label=gettext_lazy("Primary Contact"), value=True)],
    ),
    dp.Legend(
        key="company_contact",
        items=[dp.LegendItem(icon=WBColor.YELLOW_LIGHT.value, label=gettext_lazy("Company Contact"), value=True)],
    ),
]
CONTACT_PERSON_FORMATTING = [
    dp.Formatting(
        column="primary",
        formatting_rules=[
            dp.FormattingRule(
                style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                condition=("==", True),
            )
        ],
    ),
    dp.Formatting(
        column="company_contact",
        formatting_rules=[
            dp.FormattingRule(
                style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                condition=("==", True),
            )
        ],
    ),
]

CONTACT_COMPANY_LEGEND = [
    dp.Legend(
        key="primary",
        items=[dp.LegendItem(icon=WBColor.GREEN_LIGHT.value, label=gettext_lazy("Primary Contact"), value=True)],
    ),
]
CONTACT_COMPANY_FORMATTING = [
    dp.Formatting(
        column="primary",
        formatting_rules=[
            dp.FormattingRule(
                style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                condition=("==", True),
            )
        ],
    )
]


def get_legend_formatting(self):
    entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
    if entry.is_company:
        legend = CONTACT_COMPANY_LEGEND
        format = CONTACT_COMPANY_FORMATTING
        return legend, format
    else:
        legend = CONTACT_PERSON_LEGEND
        format = CONTACT_PERSON_FORMATTING
        return legend, format


class EmailContactDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="address", label=_("Email Address")),
                dp.Field(key="entry", label=_("Entry")),
            ],
            legends=CONTACT_PERSON_LEGEND,
            formatting=CONTACT_PERSON_FORMATTING,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["primary", "location"],
                ["entry", "address"],
            ]
        )


class TelephoneContactDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="telephone_type", label=_("Type")),
                dp.Field(key="number", label=_("Number")),
                dp.Field(key="entry", label=_("Entry")),
            ],
            legends=CONTACT_PERSON_LEGEND,
            formatting=CONTACT_PERSON_FORMATTING,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["primary", "location", "entry"], ["telephone_type", repeat_field(2, "number")]])


class AddressContactDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="street", label=_("Street")),
                dp.Field(key="street_additional", label=_("Street Additional")),
                dp.Field(key="zip", label=_("Zip")),
                dp.Field(key="geography_city", label=_("City")),
            ],
            legends=CONTACT_PERSON_LEGEND,
            formatting=CONTACT_PERSON_FORMATTING,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["primary", "location", "entry"],
                [repeat_field(2, "street"), "street_additional"],
                [repeat_field(2, "geography_city"), "zip"],
            ]
        )


class BankingContactDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="institute", label=_("Institute")),
                dp.Field(key="institute_additional", label=_("Institute additional")),
                dp.Field(key="iban", label=_("IBAN")),
                dp.Field(key="swift_bic", label=_("SWIFT")),
                dp.Field(key="currency", label=_("Currency")),
                dp.Field(key="entry", label=_("Entry")),
                dp.Field(key="edited", label=_("Edited")),
                dp.Field(key="additional_information", label=_("Information")),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=BankingContact.Status.APPROVED.label,
                            value=BankingContact.Status.APPROVED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=BankingContact.Status.PENDING.label,
                            value=BankingContact.Status.PENDING.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=BankingContact.Status.DRAFT.label,
                            value=BankingContact.Status.DRAFT.value,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", BankingContact.Status.APPROVED.value),
                        ),
                    ],
                ),
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", BankingContact.Status.PENDING.value),
                        )
                    ],
                ),
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", BankingContact.Status.DRAFT.value),
                        )
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        if "pk" not in self.view.kwargs:
            fields = []
        else:
            fields = [[repeat_field(2, "status")]]
        fields.extend(
            [
                [repeat_field(2, "location")],
                ["primary", "."],
                ["institute", "institute_additional"],
                ["iban", "swift_bic"],
                ["currency", "entry"],
                [repeat_field(2, "additional_information")],
                [repeat_field(2, "edited")],
            ]
        )
        return create_simple_display(fields)


class EmailContactEntryDisplay(EmailContactDisplay):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="address", label=_("Email Address")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["primary", "location"], [repeat_field(2, "address")]])


class AddressContactEntryDisplay(AddressContactDisplay):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="street", label=_("Street")),
                dp.Field(key="street_additional", label=_("Street Additional")),
                dp.Field(key="zip", label=_("Zip")),
                dp.Field(key="geography_city", label=_("City")),
                dp.Field(key="geography_state", label=_("State")),
                dp.Field(key="geography_country", label=_("Country")),
                dp.Field(key="geography_continent", label=_("Continent")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["primary", "location"],
                ["street", "street_additional"],
                ["zip", "geography_city"],
            ]
        )


class TelephoneContactEntryDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="telephone_type", label=_("Type")),
                dp.Field(key="number", label=_("Number")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["primary", "location"], ["telephone_type", "number"], [repeat_field(2, "activities_section")]],
            [
                create_simple_section(
                    "activities_section", "Activities", [["list_of_activities"]], "list_of_activities", collapsed=True
                )
            ],
        )


class WebsiteContactEntryDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="url", label=_("URL")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["primary", "location"], [repeat_field(2, "url")]])


class BankingContactEntryDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="institute", label=_("Institute")),
                dp.Field(key="institute_additional", label=_("Institute additional")),
                dp.Field(key="iban", label=_("IBAN")),
                dp.Field(key="swift_bic", label=_("SWIFT")),
                dp.Field(key="currency", label=_("Currency")),
                dp.Field(key="edited", label=_("Edited")),
                dp.Field(key="additional_information", label=_("Information")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        if "pk" not in self.view.kwargs:
            fields = []
        else:
            fields = [[repeat_field(2, "status")]]
        fields.extend(
            [
                ["primary", "location"],
                ["institute", "institute_additional"],
                ["iban", "swift_bic"],
                [repeat_field(2, "currency")],
                [repeat_field(2, "additional_information")],
            ]
        )

        return create_simple_display(fields)


class SocialMediaContactEntryDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        legend, format = get_legend_formatting(self)
        return dp.ListDisplay(
            fields=[
                dp.Field(key="platform", label=_("Platform")),
                dp.Field(key="location", label=_("Location")),
                dp.Field(key="url", label=_("URL")),
            ],
            legends=legend,
            formatting=format,
            editable=True,
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["platform", "primary"], ["url", "location"]])



from typing import Optional

from django.utils.translation import gettext as _
from django.utils.translation import pgettext

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default

transitions_section = Section(
    key="transitions_section",
    collapsible=True,
    title=_("Transitions"),
    display=Display(
        pages=[
            Page(
                title=_("Transitions"),
                layouts={
                    default(): Layout(
                        grid_template_areas=[["transitions_inline"]],
                        grid_template_columns=[
                            "minmax(min-content, 1fr)",
                        ],
                        grid_auto_rows=Style.MIN_CONTENT,
                        inlines=[Inline(key="transitions_inline", endpoint="transitions")],
                    )
                },
            ),
        ]
    ),
)
process_steps_section = Section(
    key="process_steps_section",
    collapsible=True,
    title=_("Process Steps"),
    display=Display(
        pages=[
            Page(
                title=_("Process Steps"),
                layouts={
                    default(): Layout(
                        grid_template_areas=[["process_steps_inline"]],
                        grid_template_columns=[
                            "minmax(min-content, 1fr)",
                        ],
                        grid_auto_rows=Style.MIN_CONTENT,
                        inlines=[Inline(key="process_steps_inline", endpoint="process_steps")],
                    )
                },
            ),
        ]
    ),
)


class StepDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="name", label=_("Name")),
            dp.Field(key="step_type", label=_("Step Type")),
            dp.Field(key="workflow", label=_("Workflow")),
            dp.Field(key="status", label=_("Status")),
            dp.Field(key="code", label=_("Code")),
            dp.Field(key="permission", label=_("Permission")),
        ]

        if "workflow_id" in self.view.kwargs:
            fields.pop(2)

        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "step_type", "workflow"],
            ["status", "code", "permission"],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:2]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class UserStepDisplayConfig(StepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if step_display := super().get_list_display():
            step_display.fields.pop(1)
            userstep_fields = [
                dp.Field(key="assignee", label=_("Assignee")),
                dp.Field(key="group", label=_("Group")),
                dp.Field(key="assignee_method", label=_("Assignee Method")),
                dp.Field(key="notify_user", label=_("Notify User")),
                # dp.Field(key="display", label=_("Display")),
            ]
            return dp.ListDisplay(
                fields=step_display.fields + userstep_fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "status"],
            ["code", "assignee", "notify_user"],
            ["group", "assignee_method", "permission"],
            # ["display", ".", "."],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:3]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class BaseStepDisplayConfig(StepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if step_display := super().get_list_display():
            step_display.fields.pop(1)
            return dp.ListDisplay(
                fields=step_display.fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "code"],
            ["status", "permission", "."],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:2]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class JoinStepDisplayConfig(BaseStepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if base_display := super().get_list_display():
            base_display.fields.insert(2, dp.Field(key="wait_for_all", label=_("Wait For All")))
            return dp.ListDisplay(
                fields=base_display.fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "wait_for_all"],
            ["status", "permission", "code"],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:2]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class ScriptStepDisplayConfig(BaseStepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if base_display := super().get_list_display():
            base_display.fields.insert(2, dp.Field(key="script", label=_("Script")))
            return dp.ListDisplay(
                fields=base_display.fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "code"],
            ["status", "permission", "."],
            ["script", "script", "script"],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:3]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class EmailStepDisplayConfig(BaseStepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if base_display := super().get_list_display():
            base_display.fields = (
                base_display.fields[:1]
                + [
                    dp.Field(
                        key="to",
                        label=pgettext("Email context", "To"),
                    ),
                    dp.Field(key="subject", label=_("Subject")),
                    dp.Field(key="template", label=_("Template")),
                    dp.Field(key="cc", label=_("Cc")),
                    dp.Field(key="bcc", label=_("BCC")),
                ]
                + base_display.fields[1:]
            )
            return dp.ListDisplay(
                fields=base_display.fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "code"],
            ["subject", "template", "."],
            ["to", "cc", "bcc"],
            ["status", "permission", "."],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:4]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class FinishDisplayConfig(BaseStepDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        if base_display := super().get_list_display():
            base_display.fields.insert(
                2, dp.Field(key="write_preserved_instance", label=_("Write Preserved Instance"))
            )
            return dp.ListDisplay(
                fields=base_display.fields,
            )
        return None

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow", "code"],
            ["status", "permission", "write_preserved_instance"],
            [
                "transitions_section",
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:2]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )


class StartStepDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = [
            dp.Field(key="name", label=_("Name")),
            dp.Field(key="workflow", label=_("Workflow")),
            dp.Field(key="status", label=_("Status")),
            dp.Field(key="code", label=_("Code")),
        ]

        if "workflow_id" in self.view.kwargs:
            fields.pop(1)

        return dp.ListDisplay(
            fields=fields,
        )

    def get_instance_display(self) -> Display:
        grid_fields = [
            ["name", "workflow"],
            ["status", "code"],
            [
                "transitions_section",
                "transitions_section",
            ],
            [
                "process_steps_section",
                "process_steps_section",
            ],
        ]
        if "pk" not in self.view.kwargs:
            grid_fields = grid_fields[:2]

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=grid_fields,
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[transitions_section, process_steps_section],
                        )
                    },
                ),
            ]
        )



from configurations import values


class Restframework:
    REST_FRAMEWORK_ANON_THROTTLE_RATES = values.IntegerValue(60, environ_prefix=None)
    REST_FRAMEWORK_THROTTLE_PERIOD = values.Value("hour", environ_prefix=None)
    REST_FRAMEWORK_USER_THROTTLE_RATES = values.IntegerValue(5000, environ_prefix=None)

    def REST_FRAMEWORK(self):
        rest_framework = {
            "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
            "PAGE_SIZE": 25,
            "DEFAULT_RENDERER_CLASSES": ("rest_framework.renderers.JSONRenderer",),
            "DEFAULT_AUTHENTICATION_CLASSES": (
                "rest_framework_simplejwt.authentication.JWTAuthentication",
                "wbcore.contrib.authentication.authentication.TokenAuthentication",
            ),
            "DEFAULT_PERMISSION_CLASSES": (
                "rest_framework.permissions.IsAuthenticated",
                "wbcore.permissions.permissions.RestAPIModelPermissions",
            ),
            "DATETIME_FORMAT": "%Y-%m-%dT%H:%M:%S%z",
            "DEFAULT_THROTTLE_CLASSES": [
                "rest_framework.throttling.AnonRateThrottle",
                "rest_framework.throttling.UserRateThrottle",
            ],
            "DEFAULT_THROTTLE_RATES": {
                "anon": f"{self.REST_FRAMEWORK_ANON_THROTTLE_RATES}/{self.REST_FRAMEWORK_THROTTLE_PERIOD}",
                "user": f"{self.REST_FRAMEWORK_USER_THROTTLE_RATES}/{self.REST_FRAMEWORK_THROTTLE_PERIOD}",
            },
        }
        if self.DEBUG:
            rest_framework["DEFAULT_AUTHENTICATION_CLASSES"] = (
                "rest_framework_simplejwt.authentication.JWTAuthentication",
                "wbcore.contrib.authentication.authentication.TokenAuthentication",
                "rest_framework.authentication.SessionAuthentication",
            )
            rest_framework["DEFAULT_RENDERER_CLASSES"] = (
                "rest_framework.renderers.JSONRenderer",
                "wbcore.utils.renderers.BrowsableAPIRendererWithoutForms",
            )
        return rest_framework



# Generated by Django 5.0.13 on 2025-03-10 14:40

import django.db.models.manager
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0011_person_description_person_i18n'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='person',
            managers=[
                ('all_objects', django.db.models.manager.Manager()),
            ],
        ),
    ]



from datetime import datetime
from io import BytesIO
from typing import Generator, Optional

from django.db import models


class AbstractDataBackend:
    def __init__(self, import_credential: Optional[models.Model] = None, data_backend=None, **kwargs):
        self.data_backend = data_backend
        self.import_credential = import_credential

    def is_object_valid(self, obj: models.Model) -> bool:
        """
        Can be overriden to define how an object is consired valid by the backend. Default to return True

        Args:
            obj: The tested object

        Returns:
            True if the object is considered valid
        """
        default_qs = self.get_default_queryset()
        if default_qs is not None:
            return isinstance(obj, default_qs.model) and default_qs.filter(pk=obj.pk).exists()
        return False

    def get_default_queryset(self) -> models.QuerySet | None:
        """
        Returns the defaults queryset to loop over by the backend to retreive import files. Used on daily import when no queryset is explicitly provided.

        Returns:
            The default queryset if it exists, none otherwise (Default)
        """
        return None

    def get_provider_id(self, obj: models.Model) -> str | None:
        """
        Override if backend support another and different identifier scheme. We expect this function to the external provider identifier given a certain object.

        Args:
            obj: The object to extract the external identifier

        Returns:
            The external provider identifier it it exists. None otherwise (Default)
        """
        return None

    def get_files(self, execution_time: datetime, **kwargs) -> Generator[tuple[str, BytesIO], None, None] | None:
        """
        Override to implement the import file creation logic

        Args:
            execution_time: The time at which the import takes place
            **kwargs: Possible keywords arguments to be passed down to the importer.

        Returns:
            A generator of bitestream object.
        """
        pass



from pytest_django.live_server_helper import LiveServer
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from wbcore import serializers as wb_serializers

WAIT_TIME_SEC = 30


def login(driver: WebDriver, user_mail: str | None, user_password: str | None):
    """Automatically logs in a user into the workbench.

    Args:
        driver (WebDriver): Selenium webdriver.
        user_mail (str | None): The users email.
        user_password (str | None): The users password
    """
    write_text(driver, "//input[@placeholder='E-Mail']", user_mail)
    write_text(driver, "//input[@placeholder='Password']", user_password)
    click_element_by_path(driver, "//button[@label='Login']")


def set_up(driver: WebDriver, server: LiveServer, user_mail: str, user_password: str):
    """Used for setting up the test environment and needs to be called before executing the test logic.

    Args:
        driver (WebDriver): Selenium webdriver.
        server (LiveServer): Live server.
        user_mail (str): The email of the user used for authentication.
        user_password (str): The password of the user used for authentication.
    """
    driver.get(server.url)
    driver.set_window_size(1920, 1080)
    login(driver, user_mail, user_password)


def open_menu_item(driver: WebDriver, *menu_items, perform_mouse_move=False):
    """Automatically opens a menu item of the CRM.

    Args:
        driver (WebDriver): The Selenium webdriver.
        *menu_items: The path to the menu item you want to open (e.g. for Person you would have to insert "CRM", "Persons" to get to the person list.)
        perform_mouse_move (bool, optional): Should the mouse move after opening the correct menu item. This can be helpful, when the navigation bar block certain things. When moving the mouse away from the navigation bar, the bar will collapse. Defaults to False.
    """
    actions = ActionChains(driver, 1000)
    ele = find_element(driver, "//div[@title='Navigation']")
    actions.move_to_element(ele).perform()

    for menu_item in menu_items:
        click_element_by_path(driver, f"//div[@title='{menu_item}']")

    if perform_mouse_move:
        window_width = (driver.get_window_size().get("width")) / 1.75
        actions.move_by_offset(window_width, 0).perform()
        actions.click()


def open_create_instance(driver: WebDriver, *menu_items: str, create_instance_title: str):
    """Opens a create instance widget

    Args:
        driver (WebDriver): The Selenium webdriver.
        *menu_items: The path to the menu item you want to open (e.g. for Person you would have to insert "CRM", "Persons" to get to the person list.)
        create_instance_title (str): The title of the create instance button (e.g. Create Activity)
    """
    actions = ActionChains(driver, 1000)
    open_menu_item(driver, *menu_items)
    click_element_by_path(driver, f"//span[@title='{create_instance_title}']")
    window_width = (driver.get_window_size().get("width")) / 2
    actions.move_by_offset(window_width, 0).perform()


def write_text(driver: WebDriver, xpath: str, text: str) -> WebElement:
    """Writes a given text into an element. Requires the element to be writable.

    Args:
        driver (WebDriver): Selenium webdriver.
        xpath (str): The xpath leading to the element.
        text (str): The desired text.

    Returns:
        WebElement: The element in which it was written.
    """
    element = find_element(driver, xpath)
    element.send_keys(text)

    return element


def approve_form(driver: WebDriver, element: WebElement):
    click_button_by_label(driver, "Approve")


# ====================================================================================================================================== #
# ========================================================== FINDING ELEMENTS ========================================================== #
# ====================================================================================================================================== #


def find_element(driver: WebDriver, xpath: str, wait_sec=WAIT_TIME_SEC) -> WebElement | None:
    """Find a single web element on the current page.

    Args:
        driver (WebDriver): Selenium webdriver.
        xpath (str): The xpath used to look up the element.
        wait_sec (int): the seconds the driver waits for the element to be visible.

    Returns:
        WebElement: The found element.
    """
    try:
        return WebDriverWait(driver, wait_sec).until(lambda x: x.find_element(By.XPATH, xpath))
    except TimeoutException:
        return None


def find_element_by_text(driver: WebDriver, text: str, wait_sec=WAIT_TIME_SEC) -> WebElement | None:
    """Find a web element based on a string.

    Args:
        driver (WebDriver): Selenium webdriver.
        text (str): The text that appears in the element we are searching

    Returns:
        WebElement: The found element.
    """
    try:
        xpath: str = f"//*[contains(text(), '{text}')]//parent::div"
        return WebDriverWait(driver, wait_sec).until(lambda x: x.find_element(By.XPATH, xpath))
    except TimeoutException:
        return None


# ====================================================================================================================================== #
# ========================================================= CLICKING ELEMENTS ========================================================== #
# ====================================================================================================================================== #


def click_element(driver: WebDriver, element: WebElement) -> None:
    """Click a given web element.

    Args:
        driver (WebDriver): Selenium webdriver.
        xpath (str): The element that needs clicking.
    """
    element.click()


def click_element_by_path(driver: WebDriver, xpath: str, wait_sec=WAIT_TIME_SEC) -> None:
    """Click on the item corresponding to the xpath.

    Args:
        driver (WebDriver): Selenium webdriver.
        xpath (str): The xpath leading to the element.
    """
    try:
        WebDriverWait(driver, wait_sec).until(EC.presence_of_element_located((By.XPATH, xpath))).click()
    except TimeoutException:
        return []


def click_button_by_label(driver: WebDriver, label: str):
    """Clicks a button according to its label.

    Args:
        driver (WebDriver): Selenium webdriver.
        label (str): The button label.
    """
    click_element_by_path(driver, f"//button[@label='{label}']")


def click_new_button(driver: WebDriver):
    """Clicks the new button on the current page.

    Args:
        driver (WebDriver): Selenium webdriver.
    """
    click_element_by_path(driver, "//button[contains(@class, 'new-button')]")


def click_cancel_on_button(driver: WebDriver, button_label: str):
    """Clicks the cancel symbol on the button with the specified label.

    Args:
        driver (WebDriver): Selenium webdriver.
        label (str): The button label.
    """
    click_element_by_path(driver, f"//td[@title='{button_label}']//parent::tr//div//span[.='cancel']")


def click_close_all_widgets(driver: WebDriver):
    """Clicks the close all widgets button.

    Args:
        driver (WebDriver): Selenium webdriver.
    """
    click_element_by_path(driver, "//button[@title='Close all widgets']")


# ====================================================================================================================================== #
# ========================================================= HANDLE FORM FIELDS ========================================================= #
# ====================================================================================================================================== #


def edit_form_fields(driver: WebDriver, serializer: wb_serializers, changes: dict[str]):
    """A method to edit the fields in a form.

    Args:
        driver (WebDriver): The Selenium webdriver.
        serializer (wb_serializers): The serializer that matches the model of the form.
        changes (dict[str]): A dict containing the changes. The dict keys should have the names of the corresponding fields (e.g. {"last_name": "Hotzenplotz"}).
    """
    field_list = list(changes.keys())
    for field in field_list:
        if not changes.get(field):
            continue
        if type(serializer.fields[field]) is wb_serializers.CharField:
            find_element(
                driver, f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//input"
            ).clear()
            write_text(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//input",
                changes[field],
            )
        elif type(serializer.fields[field]) is wb_serializers.ChoiceField:
            click_element_by_path(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//span[text()='expand_more']",
            )
            click_element_by_path(driver, f"//div[contains(text(),'{changes[field]}')]")
        elif type(serializer.fields[field]) is wb_serializers.PrimaryKeyRelatedField:
            click_element_by_path(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//span[text()='expand_more']",
            )
            click_element_by_path(driver, f"//li[contains(text(),'{changes[field]}')]")


def fill_out_form_fields(driver: WebDriver, serializer: wb_serializers, field_list: list[str], entity):
    """A method to fill out the fields in a form.

    Args:
        driver (WebDriver): The Selenium webdriver.
        serializer (wb_serializers): The serializer that matches the model of the form.
        field_list (list[str]): A list of the field names to be changed.
        entity (_type_): An entity on which the changes to the form are based. The entity should be created via the corresponding factory. (The entity should be created by using the Factory build method: e.g. PersonFactory.build())
    """
    for field in field_list:
        if type(serializer.fields[field]) in [wb_serializers.CharField, wb_serializers.IntegerField]:
            write_text(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//input",
                getattr(entity, field),
            )
        elif type(serializer.fields[field]) is wb_serializers.ChoiceField:
            click_element_by_path(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//span[text()='expand_more']",
            )
            click_element_by_path(driver, f"//div[contains(text(), '{getattr(entity, field)}')]")
        elif type(serializer.fields[field]) is wb_serializers.PrimaryKeyRelatedField:
            click_element_by_path(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//span[text()='expand_more']",
            )
            click_element_by_path(driver, f"//li[contains(text(), '{getattr(entity, field)}')]")
        elif type(serializer.fields[field]) is wb_serializers.BooleanField:
            click_element_by_path(
                driver,
                f"//label[contains(text(), '{serializer.fields[field].label}')]//parent::div//div//span[text()='{'Yes' if {getattr(entity, field)} else 'No'}']",
            )


# ====================================================================================================================================== #
# ======================================================== HANDLE LIST INSTANCES ======================================================= #
# ====================================================================================================================================== #


def edit_list_instance(
    driver: WebDriver,
    actions,
    list_name: str,
    serializer: wb_serializers.ModelSerializer,
    new_inputs: dict,
):
    """Edits a list instance.

    Args:
        driver (WebDriver): The Selenium webdriver.
        actions (_type_): The selenium action chains.
        list_name (str): The name of the list (e.g.: Persons)
        serializer (wb_serializers.ModelSerializer): The model serializer that matches the lists model.
        new_inputs (dict): A dict containing changes. The dict keys should have the names of the corresponding fields (e.g. {"last_name": "Hotzenplotz"}).
    """
    list_element = find_element(driver, f"//div[text()='{list_name}']")
    actions.double_click(list_element).perform()
    if unlock_button := find_element(driver, "//button[@label='Unlock']"):
        click_element(driver, unlock_button)
    edit_form_fields(driver, serializer, new_inputs)
    click_element_by_path(driver, "//button[@label='Save and close']")


def delete_list_entry(driver: WebDriver, actions: ActionChains, title: str):
    """Deletes a list entry. Requires that you are already in the correct list view.

    Args:
        driver (WebDriver): The Selenium webdriver.
        actions (ActionChains): The selenium action chains.
        title (str): The title of the list element that shall be deleted.
    """
    list_element = find_element(driver, f"//div[text()='{title}']")
    actions.context_click(list_element).perform()
    click_element_by_path(driver, "//span[text()='Delete instance']")
    click_element_by_path(driver, "//button[@label='Confirm']")


# ====================================================================================================================================== #
# =========================================================== HANDLE FILTERS =========================================================== #
# ====================================================================================================================================== #


def navigate_to_filter(driver: WebDriver, *menu_items):
    """Navigates to a specified list instance and opens the filter menu there.

    Args:
        driver (WebDriver): Selenium webdriver
        menu_item (str): The menu in which the filters should be open. You need to specify the whole path to the menu. E.g. ["CRM", "Persons"] to get to the CRMs persons menu.
    """
    actions = ActionChains(driver, 1000)
    open_menu_item(driver, *menu_items)
    # we need to move the mouse away from the side bar, so that the side bar closes.
    window_width = (driver.get_window_size().get("width")) / 2
    window_height = (driver.get_window_size().get("height")) / 2
    actions.move_by_offset(window_width, window_height).perform()

    filter_menu_btn = find_element(driver, "//button[@data-testid='control-bar-toggle-button']")
    click_element(driver, filter_menu_btn)


def select_filter(driver: WebDriver, label: str, option: str):
    """Select a filter from the filter menu a filter by the corresponding label.

    Args:
        driver (WebDriver): Selenium webdriver
        label (str): The filter label. (e.g. "Status" for the status - filter )
        option (str): The filter choice. (e.g. the "Planned" choice for the status filter)
    """
    no_activity_choices = find_element(
        driver, f"//label[text()='{label}']/parent::div//parent::div/button[contains(span, 'expand_more')]"
    )
    click_element(driver, no_activity_choices)
    click_element_by_path(driver, f"//li//div[text()='{option}']")
    click_element_by_path(driver, "//button[@label='Apply']")


def select_async_filter(driver: WebDriver, label: str, option: str):
    """Select a async filter from the filter menu a filter by the corresponding label.

    Args:
        driver (WebDriver): Selenium webdriver
        label (str): The filter label. (e.g. "Status" for the status - filter )
        option (str): The filter choice. (e.g. the "Planned" choice for the status filter)
    """
    no_activity_choices = find_element(
        driver, f"//label[text()='{label}']/parent::div//parent::span[contains(text(), 'expand_more')]"
    )
    click_element(driver, no_activity_choices)
    click_element_by_path(driver, f"//li[@title='{option}']")
    click_element_by_path(driver, "//button[@label='Apply']")


def select_column_filter(driver: WebDriver, label: str, option: str):
    """Select a filter from a list column by the corresponding label.

    Args:
        driver (WebDriver): Selenium webdriver
        label (str): The filter label. (e.g. "Status" for the status - filter )
        option (str): The filter choice. (e.g. the "Planned" choice for the status filter)
    """
    click_element_by_path(
        driver,
        f"//div[contains(@class, 'ag-cell-label-container') and ./div/span/text()='{label}']//span[contains(@class, 'ag-header-cell-menu-button')]",
    )
    click_element_by_path(
        driver, "//div[@class='ag-tabs-header ag-menu-header']//span[contains(@class, 'ag-icon-filter')]"
    )
    click_element_by_path(
        driver,
        f"//div[contains(@class, 'ag-virtual-list-container')]//div[contains(@class, 'ag-virtual-list-item') and ./div/div/div/text()='{option}']//input",
    )
    click_element_by_path(driver, "//button[@type='submit']")



from typing import Optional

from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class UtilsDisplayMixin(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(fields=[dp.Field(key="title", label=_("Title"))])

    def get_instance_display(self) -> Display:
        return create_simple_display([["title"]])


class CustomerStatusDisplay(UtilsDisplayMixin):
    pass


class PositionDisplay(UtilsDisplayMixin):
    pass


class CompanyTypeDisplay(UtilsDisplayMixin):
    pass


class SpecializationDisplay(UtilsDisplayMixin):
    pass



from dataclasses import dataclass


@dataclass
class Person:
    first_name: str
    last_name: str
    email: str
    id: str = None

    def __eq__(self, other):
        if other and (
            (self.id and other.id and self.id == other.id)
            or (self.email and other.email and self.email == other.email)
        ):
            return True
        return super().__eq__(other)



import urllib.parse
from typing import Dict, Tuple

from django.conf import settings
from django.contrib.sites.models import Site
from django.http.request import QueryDict
from django.utils.http import urlencode


def clean_shareable_url(url):
    endpoint = url.split("?widget_endpoint=")[-1]
    parse_url = urllib.parse.urlsplit(urllib.parse.unquote(endpoint))
    endpoint = parse_url.path
    if params := dict(urllib.parse.parse_qsl(parse_url.query)):
        params_repr = "&".join([f"{k}={v}" for k, v in params.items()])
        endpoint += f"?{params_repr}"
    return endpoint


def get_parse_endpoint(endpoint: str) -> Tuple[str, Dict[str, str]]:
    """
    Takes a URL endpoint and outputs the path and the params separately.

    Parameters
    ----------
    endpoint: The endpoint string

    Returns
    -------
    Returns a tuple in the form of (path: str, params: dict[str, str]). Where path is the url and params are the parameters of the endpoint.
    """

    path = urllib.parse.urljoin(endpoint, urllib.parse.urlparse(endpoint).path)
    params = dict(urllib.parse.parse_qsl(urllib.parse.urlsplit(endpoint).query))
    return path, params


def get_urlencode_endpoint(endpoint: str, params: Dict[str, str]) -> str:
    """
    Takes a URL path and params and creates an endpoint from it.

    Parameters
    ----------
    endpoint: The URL string
    params: The params dictonary

    Returns
    -------
    Returns an endpoint string with the params attached.
    """
    if isinstance(params, QueryDict):
        params = params.dict()
    return f"{endpoint}?{urlencode(params)}"


def base_domain() -> str:
    """A utility method that assembles the current domain. Utilizes the site app from django

    Returns:
        A string containing the current domain as noted in the curret `Site` prefixed with the http scheme

    """
    scheme = "https" if settings.SECURE_SSL_REDIRECT else "http"
    base_domain = Site.objects.get_current().domain
    return f"{scheme}://{base_domain}"


def new_mode(url: str) -> str:
    operator = "&" if "?" in url else "?"
    return url + f"{operator}new_mode=true"



import pytest

from wbcore.contrib.notifications.backends.firebase.backends import (
    NotificationBackend as FirebaseNotificationBackend,
)
from wbcore.contrib.notifications.factories.tokens import (
    NotificationUserTokenModelFactory,
)
from wbcore.contrib.notifications.models.tokens import NotificationUserToken


@pytest.mark.django_db
def test_send_notification(mocker, notification, user):
    mocked_get_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.get_app",
    )
    mocked_initialize_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.initialize_app",
    )
    mocked_send = mocker.patch("wbcore.contrib.notifications.backends.firebase.backends.messaging.send")
    mocked_get_firebase_credentials = mocker.patch.object(FirebaseNotificationBackend, "get_firebase_credentials")

    NotificationUserTokenModelFactory.create(user=user, device_type=NotificationUserToken.NotificationDeviceType.WEB)
    notification.notification_type.user_settings.filter(user=notification.user).update(enable_web=True)
    FirebaseNotificationBackend.send_notification(notification)

    mocked_send.assert_called_once()
    mocked_get_firebase_credentials.assert_called_once()
    mocked_get_app.assert_called_once()
    mocked_initialize_app.assert_not_called()


@pytest.mark.django_db
def test_send_multiple_notifications(mocker, notification, user):
    mocked_get_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.get_app",
    )
    mocked_initialize_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.initialize_app",
    )
    mocked_send = mocker.patch("wbcore.contrib.notifications.backends.firebase.backends.messaging.send")
    mocked_get_firebase_credentials = mocker.patch.object(FirebaseNotificationBackend, "get_firebase_credentials")

    NotificationUserTokenModelFactory.create_batch(
        3, user=user, device_type=NotificationUserToken.NotificationDeviceType.WEB
    )
    notification.notification_type.user_settings.filter(user=notification.user).update(enable_web=True)
    FirebaseNotificationBackend.send_notification(notification)

    assert mocked_send.call_count == 3
    mocked_get_firebase_credentials.assert_called_once()
    mocked_get_app.assert_called_once()
    mocked_initialize_app.assert_not_called()


@pytest.mark.django_db
def test_send_notification_with_initialize(mocker, notification, user):
    mocked_get_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.get_app",
        side_effect=ValueError(),
    )
    mocked_initialize_app = mocker.patch(
        "wbcore.contrib.notifications.backends.firebase.backends.firebase_admin.initialize_app"
    )
    mocked_send = mocker.patch("wbcore.contrib.notifications.backends.firebase.backends.messaging.send")
    mocked_get_firebase_credentials = mocker.patch.object(FirebaseNotificationBackend, "get_firebase_credentials")
    NotificationUserTokenModelFactory.create(user=user, device_type=NotificationUserToken.NotificationDeviceType.WEB)
    notification.notification_type.user_settings.filter(user=notification.user).update(enable_web=True)
    FirebaseNotificationBackend.send_notification(notification)

    mocked_get_app.assert_called_once()
    mocked_initialize_app.assert_called_once()
    mocked_send.assert_called_once()
    mocked_get_firebase_credentials.assert_called_once()


def test_get_configuration(mocker):
    mocker.patch.dict("os.environ", {"FIREBASE_WEB_CONFIG": '{"ABC": "123"}', "FIREBASE_VAPID_KEY": "DEF"})
    assert FirebaseNotificationBackend.get_configuration() == {"firebase_config": {"ABC": "123"}, "vapid_key": "DEF"}



from .event import EventFactory, EventTypeFactory
from .league import LeagueFactory
from .match import MatchFactory
from .person import SportPersonFactory, PlayerFactory
from .role import RoleFactory
from .sport import SportFactory
from .stadium import StadiumFactory
from .team import TeamFactory, TeamResultsFactory



from . import AbstractBackend


class IconBackend(AbstractBackend):
    """
    List of Icon supported from the Google Icon Material.

    They can be added and search through https://fonts.google.com/icons?icon.set=Material+Icons
    """

    @classmethod
    @property
    def fallback_icon(cls):
        return "broken_image"

    ACCOUNT_BALANCE = "account_balance"
    ADD = "add_circle"
    APPROVE = "thumb_up_alt"
    BANK = "account_balance"
    BIRTHDAY = "cake"
    BOOKMARK = "bookmark_border"
    CALENDAR = "calendar_month"
    CHART_AREA = "stacked_line_chart"
    CHART_BARS_HORIZONTAL = "align_horizontal_center"
    CHART_BARS_VERTICAL = "bar_chart"
    CHART_BARS_VERTICAL_INCREASING = "signal_cellular_alt"
    CHART_LINE = "show_chart"
    CHART_PIE = "pie_chart"
    CHART_PYRAMID = "bar_chart"
    CHART_SWITCHES = "candlestick_chart"
    CHART_WATERFALL = "waterfall_chart"
    CHART_MULTILINE = "ssid_chart"
    CLIPBOARD = "assignment"
    CONFIGURE = "tune"
    CONFIRM = "check_circle"
    COPY = "content_copy"
    CURRENCY_EXCHANGE = "currency_exchange"
    DASHBOARD = "dashboard"
    DATA_GRID = "analytics"
    DATA_LIST = "table_view"
    DATA_EXPLORATION = "query_stats"
    DAY_OFF = "event_busy"
    DEAL = "handshake"
    DEAL_MONEY = "request_quote"
    DECISION_STEP = "alt_route"
    DELETE = "delete"
    DENY = "thumb_down_alt"
    DISABLED = "block"
    DOCUMENT = "description"
    DOCUMENT_IN_PROGRESS = "history_edu"
    DOCUMENT_PRIVATE = "assignment_late"
    DOCUMENT_WITH_DOLLAR = "request_page"
    DOCUMENT_WITH_ATTACHMENT = "file_present"
    DOLLAR = "attach_money"
    DOWNLOAD = "file_download"
    EDIT = "edit"
    END = "last_page"
    ENUMERATION = "format_list_numbered"
    EURO = "euro"
    EVENT = "event"
    EVENT_ACCEPTED = "event_available"
    EXPEND = "fullscreen"
    FAVORITE = "star"
    FEEDBACK = "feedback"
    FOLDERS_ADD = "create_new_folder"
    FOLDERS_MONEY = "folder_special"
    FOLDERS_OPEN = "folder_open"
    GENERATE_NEXT = "next_plan"
    GOAL = "sports_soccer"
    GROUPS = "groups"
    HOME = "home"
    HISTORY = "history"
    IGNORE = "visibility_off"
    IMPORT_EXPORT = "import_export"
    INFO = "info"
    INVERSE = "swap_horiz"
    JOIN_STEP = "merge"
    LINK = "link"
    LOCATION = "location_on"
    LOCK = "lock"
    LUNCH = "restaurant"
    MAIL = "email"
    MAIL_MONEY = "attach_email"
    MAIL_OPEN = "mark_email_read"
    MAIL_POSTED = "forward_to_inbox"
    MERGE = "merge"
    NEWSPAPER = "newspaper"
    NEXT = "fast_forward"
    NOTEBOOK = "notes"
    OTHER = "more_horiz"
    PENDING = "pending"
    PEOPLE = "people_alt"
    PERSON = "person"
    PERSON_ADD = "person_add"
    PHONE = "smartphone"
    PHONE_ADD = "add_ic_call"
    PREVIOUS = "fast_rewind"
    PUBLIC = "public"
    QUESTION = "question_mark"
    REDO = "redo"
    REFRESH = "refresh"
    REGENERATE = "restart_alt"
    REJECT = "cancel"
    REPEAT = "repeat"
    REPLACE = "published_with_changes"
    REVIEW = "rate_review"
    RUNNING = "running_with_errors"
    SAVE = "save"
    SCHEDULE = "schedule"
    SCRIPT = "terminal"
    SEND = "send"
    SEND_LATER = "schedule_send"
    SETTINGS = "settings"
    SHARE = "share"
    SHRINK = "fullscreen_exit"
    SICK_LEAVE = "sick"
    SPLIT = "call_split"
    START = "start"
    STATS = "query_stats"
    SUPERVISE = "supervised_user_circle"
    SYNCHRONIZE = "sync"
    TABLE = "table_chart"
    TIME_UP = "hourglass_bottom"
    TRADE = "request_quote"
    TRAINING = "school"
    UNDO = "restore"
    UNFILTER = "filter_alt_off"
    UNFOLD = "arrow_drop_down"
    UNLINK = "link_off"
    UNLOCK = "lock_open"
    UPLOAD = "file_upload"
    VACATION = "holiday_village"
    VIEW = "visibility"
    WARNING = "warning"
    WORK = "card_travel"



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class GeographyDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="parent", label="Parent"),
                dp.Field(key="code_2", label="Alpha 2 Country Code"),
                dp.Field(key="code_3", label="Alpha 3 Country Code"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [
                    "name",
                    "parent",
                    "code_2",
                    "code_3",
                ]
            ]
        )



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

YOUR_CONTACT_MENUITEM = MenuItem(
    label=_("Your Contact"),
    endpoint="wbcore:directory:clientmanagerrelationship-userclient-list",
    permission=ItemPermission(method=lambda request: request.user.is_active and not request.user.profile.is_internal),
)

COMPANY_MENUITEM = MenuItem(
    label=_("Companies"),
    endpoint="wbcore:directory:company-list",
    permission=ItemPermission(permissions=["directory.view_company"]),
    add=MenuItem(
        label=_("Create Company"),
        endpoint="wbcore:directory:company-list",
        permission=ItemPermission(permissions=["directory.add_company"]),
    ),
)
PERSON_MENUITEM = MenuItem(
    label=_("Persons"),
    endpoint="wbcore:directory:person-list",
    permission=ItemPermission(permissions=["directory.view_person"]),
    add=MenuItem(
        label=_("Create Person"),
        endpoint="wbcore:directory:person-list",
        permission=ItemPermission(permissions=["directory.add_person"]),
    ),
)

SYSTEMEMPLOYEE_MENUITEM = MenuItem(
    label=_("Employees"),
    endpoint="wbcore:directory:systememployee-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["directory.view_person"]
    ),
)

BANK_MENUITEM = MenuItem(
    label=_("Banks"),
    endpoint="wbcore:directory:bank-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["directory.view_bank"]
    ),
    add=MenuItem(
        label=_("Create Bank"),
        endpoint="wbcore:directory:bank-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user), permissions=["directory.add_bank"]
        ),
    ),
)

USERISMANAGER_MENUITEM = MenuItem(
    label=_("Your Clients/Prospects/Contacts"),
    endpoint="wbcore:directory:clientmanagerrelationship-usermanager-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["directory.view_person"]
    ),
)



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class BuildingTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Building: {{name}}")

    def get_list_title(self):
        return _("Buildings")

    def get_create_title(self):
        return _("New Building")


class ConferenceRoomTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Conference Room: {{name}}")

    def get_list_title(self):
        return _("Conference Rooms")

    def get_create_title(self):
        return _("New Conference Room")



from calendar import monthrange
from datetime import date, timedelta

from django.utils.timezone import localdate
from psycopg.types.range import DateRange
from rest_framework.exceptions import APIException

from wbcore.utils.date import get_end_date_from_date, get_start_date_from_date


def current_quarter_date_start(*args, **kwargs):
    return get_start_date_from_date(localdate())


def current_quarter_date_end(*args, **kwargs):
    return get_end_date_from_date(localdate())


def current_quarter_date_range(*args, **kwargs):
    return DateRange(current_quarter_date_start(*args, **kwargs), current_quarter_date_end(*args, **kwargs))


def current_month_date_start(*args, **kwargs):
    today = date.today()
    return date(today.year, today.month, 1)


def current_month_date_end(*args, **kwargs):
    today = date.today()
    return date(today.year, today.month, monthrange(today.year, today.month)[1])


def current_month_date_range(*args, **kwargs):
    return DateRange(current_month_date_start(*args, **kwargs), current_month_date_end(*args, **kwargs))


def current_year_date_range(*args, **kwargs):
    d = localdate()
    return DateRange(date(d.year, 1, 1), max((date(d.year + 1, 1, 1), d + timedelta(days=60))))


def year_to_date_range(*args, **kwargs):
    d = localdate()
    return DateRange(d.replace(day=1, month=1), d.today())


def n_year_data_range(n: int, *args, **kwargs):
    d = localdate()
    return DateRange(d - timedelta(days=n * 365), d)


def one_year_date_range(*args, **kwargs):
    return n_year_data_range(1)


def five_year_data_range(*args, **kwargs):
    return n_year_data_range(5)


def current_quarter_date_interval(*args, **kwargs):
    return (
        current_quarter_date_start(),
        current_quarter_date_end(*args, **kwargs),
    )


class RequiredFilterMissing(APIException):
    status_code = 400
    default_detail = "Required filters are not available in the query parameters."
    default_code = "required_filters_missing"



from typing import Callable

import pytest

from wbcore.contrib.geography.models import Geography


class TestModels:
    @pytest.fixture
    def geography(self):
        return Geography(name="Test Geography")

    def test_endpoint_basename(self, geography):
        assert isinstance(geography.get_endpoint_basename, Callable)

    def test_representation_endpoint(self, geography):
        assert isinstance(geography.get_representation_endpoint, Callable)

    def test_representation_value_key(self, geography):
        assert geography.get_representation_value_key() == "id"

    def test_representation_label_key(self, geography):
        assert geography.get_representation_label_key is not None

    def test_str(self, geography):
        assert isinstance(str(geography), str)
        assert str(geography) == geography.name



# Generated by Django 4.1.10 on 2023-08-08 13:31

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("io", "0001_initial_squashed"),
    ]

    operations = [
        migrations.AddField(
            model_name="importsource",
            name="creator",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
                verbose_name="Creator",
            ),
        ),
    ]



from .utils import PreAction
from .viewsets import (
    ChartViewSet,
    TimeSeriesChartViewSet,
    HTMLViewSet,
    InfiniteDataModelView,
    ModelViewSet,
    ReadOnlyInfiniteModelViewSet,
    ReadOnlyModelViewSet,
    RepresentationViewSet,
    ViewSet,
)



import urllib
from datetime import datetime, timedelta

import pytest
from django.urls import resolve
from django.utils import timezone
from faker import Faker
from rest_framework import exceptions
from rest_framework.test import APIRequestFactory

from wbcore.contrib.authentication.authentication import inject_short_lived_token
from wbcore.contrib.authentication.models import Token

fake = Faker()
now = datetime.now()


class TestTokenUnitTests:
    @pytest.fixture
    def request_with_user(user):
        request = APIRequestFactory().get("/")
        request.user = user
        return request

    def setup_method(self):
        self.token = Token()

    @pytest.mark.parametrize("is_valid", [True, False])
    def test_check_validity_for_endpoint_for_infinite_token(self, is_valid, mocker):
        mocker.patch.object(self.token, "is_valid", is_valid)
        if self.token.is_valid:
            assert self.token.check_validity_for_endpoint(fake.sentence(), fake.date_time())
        else:
            with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
                assert self.token.check_validity_for_endpoint(fake.sentence(), fake.date_time())
            assert str(exc_info.value) == "Token is not valid anymore"

    def test_check_invalidity_for_endpoint_for_expired_token(self, mocker):
        mocker.patch.object(self.token, "valid_until", now)
        assert self.token.check_validity_for_endpoint(
            self.token.protected_view_name, self.token.valid_until - timedelta(minutes=1)
        )
        with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
            self.token.check_validity_for_endpoint(
                self.token.protected_view_name, self.token.valid_until + timedelta(minutes=1)
            )
        assert str(exc_info.value) == "Token has expired"

    def test_check_invalidity_for_endpoint_for_exhausted_token(self, mocker):
        mocker.patch.object(self.token, "number_usage_left", 2)
        mocker.patch.object(self.token, "save")
        while self.token.number_usage_left > 0:
            assert self.token.check_validity_for_endpoint(self.token.protected_view_name, fake.date_time())
        with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
            self.token.check_validity_for_endpoint(self.token.protected_view_name, fake.date_time())
        assert str(exc_info.value) == "Token usage exhausted"

    def test_check_unvalidity_for_endpoint_for_wrong_endpoint(self, mocker):
        mocker.patch.object(self.token, "protected_view_name", "Some View Name")
        with pytest.raises(exceptions.AuthenticationFailed) as exc_info:
            self.token.check_validity_for_endpoint("wbcore:authentication:wrongview", fake.date_time())
        assert str(exc_info.value) == "Token has limited scope that doesn't match the requested endpoint"

    @pytest.mark.django_db
    @pytest.mark.parametrize("view_name", [(None, "wbcore:authentication:user")])
    def test_generate_token_for_user(self, user, view_name):
        d1 = fake.date_time()
        n1 = fake.pyint()
        token = Token.generate_token_for_user(user, protected_view_name=view_name, valid_until=d1, number_of_usages=n1)
        assert token.user == user
        assert token.protected_view_name == view_name
        assert token.valid_until == d1
        assert token.number_usage_left == n1
        token.is_valid = False
        token.save()  # invalid token to check that retrieving it will enable it

        d2 = fake.date_time()
        n2 = fake.pyint()
        new_token = Token.generate_token_for_user(
            user, protected_view_name=view_name, valid_until=d2, number_of_usages=n2
        )
        assert new_token == token  # check that the uniqueness on schema work
        assert new_token.valid_until == d2
        assert new_token.number_usage_left == n2
        assert new_token.is_valid is True

    @pytest.mark.django_db
    @pytest.mark.parametrize(
        "endpoint,creation_time",
        [
            ("http://localhost:5000/wbcore/authentication/user/", now),
            ("http://localhost:5000/wbcore/authentication/user/?some_param=True", now),
            ("/wbcore/authentication/user/?some_param=True&other_param=foo", now),
        ],
    )
    def test_generate_shareable_link(self, user, endpoint, creation_time):
        creation_time = timezone.make_aware(creation_time)
        parse_url = urllib.parse.urlsplit(urllib.parse.unquote(endpoint))
        base_params = dict(urllib.parse.parse_qsl(parse_url.query))
        link = self.token.generate_shareable_link(endpoint, user, creation_time=creation_time)
        res_params = dict(urllib.parse.parse_qsl(urllib.parse.urlsplit(urllib.parse.unquote(link)).query))

        assert endpoint in link  # check that the resulting endpoint is not modified
        for k, v in base_params.items():  # check that the base query parameters weren't touched
            assert res_params[k] == v
        assert (
            res_params["token"]
            == Token.objects.get(user=user, protected_view_name=resolve(parse_url.path).view_name).key
        )  # check that the token was injected as query parameter

    @pytest.mark.django_db
    @pytest.mark.parametrize("view_name", [("wbcore:authentication:user", None)])
    def test_inject_short_lived_token(self, user, view_name, request_with_user):
        link = fake.url() + "?a=a&b=b"

        @inject_short_lived_token(view_name=view_name)
        def dummy_resources(serializer, instance, request, user, **kwargs):
            return {"resource": link}

        res = dummy_resources(None, None, request_with_user, user)
        res_params = dict(urllib.parse.parse_qsl(urllib.parse.urlsplit(urllib.parse.unquote(res["resource"])).query))
        assert link in res["resource"]  # check that the resulting endpoint is not modified
        assert res_params["a"] == "a"
        assert res_params["b"] == "b"
        assert (
            res_params["token"] == Token.objects.get(user=user, protected_view_name=view_name).key
        )  # check that the token was injected as query parameter



from rest_framework.test import APIRequestFactory

from wbcore.serializers import AdditionalResourcesField, HyperlinkField


class TestAdditionalResourcesField:
    def setup_method(self):
        self.factory = APIRequestFactory()
        self.field = AdditionalResourcesField()

    def test_not_none(self):
        assert self.field is not None


class TestHyperlinkFieldField:
    def setup_method(self):
        self.factory = APIRequestFactory()
        self.field = HyperlinkField(reverse_name="wbcore:notification-list")

    def test_not_none(self):
        assert self.field is not None

    def test_to_representation(self):
        assert self.field.to_representation("abc") == "abc"
